"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from copy import deepcopy

import ANNarchy

from ANNarchy.generator.Template.GlobalOperationTemplate import global_operation_templates_st_extern as global_op_extern_dict
from ANNarchy.generator.Utils import generate_equation_code, tabify, remove_trailing_spaces
from ANNarchy.core import Global
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

from ANNarchy.generator.Population.PopulationGenerator import PopulationGenerator
from ANNarchy.generator.Population.SingleThreadTemplates import single_thread_templates

class SingleThreadGenerator(PopulationGenerator):
    """
    Generate the header for a Population object to run either on single core
    or multi-cores with OpenMP.
    """
    def __init__(self, profile_generator, net_id):
        super(SingleThreadGenerator, self).__init__(profile_generator, net_id)

    ##################################################
    # Main method
    ##################################################
    def header_struct(self, pop, annarchy_dir):
        """
        Specialized implementation of PopulationGenerator.header_struct() for
        generation of an openMP header.

        two passes:

            * generate the codes for population header
            * fill the dictionary with call codes (return)
        """
        self._templates = deepcopy(single_thread_templates)

        # Generate declaration and accessors of all parameters and variables
        declaration_parameters_variables, access_parameters_variables = self._generate_decl_and_acc(pop)

        # Additional includes and structures
        include_additional = ""
        access_additional = ""
        struct_additional = ""
        declare_additional = ""
        init_additional = ""
        reset_additional = ""

        # Declare global operations as extern at the beginning of the file
        extern_global_operations = ""
        for op in pop.global_operations:
            extern_global_operations += global_op_extern_dict[op['function']] % {'type': ConfigManager().get('precision', self._net_id)}

        # Initialize parameters and variables
        init_parameters_variables = self._init_population(pop)

        # Spike-specific stuff
        reset_spike = ""; declare_spike = ""; init_spike = ""
        if pop.neuron_type.description['type'] == 'spike':
            spike_specific_tpl = self._templates['spike_specific']

            # Main data for spiking pops
            declare_spike += spike_specific_tpl['spike']['declare'] % {'id': pop.id}
            init_spike += spike_specific_tpl['spike']['init'] % {'id': pop.id}
            reset_spike += spike_specific_tpl['spike']['reset'] % {'id': pop.id}

            # If there is a refractory period
            if pop.neuron_type.refractory or pop.refractory:
                declare_spike += spike_specific_tpl['refractory']['declare'] % {'id': pop.id}
                if isinstance(pop.neuron_type.description['refractory'], str): # no need to instantiate refractory
                    init_spike += spike_specific_tpl['refractory']['init_extern'] % {'id': pop.id}
                else:
                    init_spike += spike_specific_tpl['refractory']['init'] % {'id': pop.id}
                reset_spike += spike_specific_tpl['refractory']['reset'] % {'id': pop.id}

            # If axonal spike condition was defined
            if pop.neuron_type.axon_spike:
                declare_spike += spike_specific_tpl['axon_spike']['declare']
                init_spike += spike_specific_tpl['axon_spike']['init']
                reset_spike += spike_specific_tpl['axon_spike']['reset']

        # Process eventual delay
        declare_delay = ""; init_delay = ""; update_delay = ""; update_max_delay = ""; reset_delay = ""
        if pop.max_delay > 1:
            declare_delay, init_delay, update_delay, update_max_delay, reset_delay = self._delay_code(pop)

        # Process mean FR computations
        declare_FR, init_FR, reset_FR = self._init_fr(pop)
        reset_spike += reset_FR

        # Update random distributions
        update_rng = self._update_random_distributions(pop)

        # Update global operations
        update_global_ops = self._update_globalops(pop)

        # Defintion of local functions
        declaration_parameters_variables += self._local_functions(pop)

        # Update the neural variables
        if pop.neuron_type.type == 'rate':
            update_variables = self._update_rate_neuron(pop)
            test_spike_cond = ""
        else:
            update_variables = self._update_spiking_neuron(pop)
            test_spike_cond = self._spike_gather(pop)

        # Stop condition
        stop_condition = self._stop_condition(pop)

        # Memory management
        size_in_bytes = self._size_in_bytes(pop)
        clear_container = self._clear_container(pop)

        # Profiling
        if self._prof_gen:
            include_profile = self._prof_gen.generate_include()
            declare_profile, init_profile = self._prof_gen.generate_init_population(pop)
        else:
            include_profile = ""
            init_profile = ""
            declare_profile = ""

        ## When everything is generated, we override the fields defined by the specific population
        if 'include_additional' in pop._specific_template.keys():
            include_additional = pop._specific_template['include_additional']
        if 'struct_additional' in pop._specific_template.keys():
            struct_additional = pop._specific_template['struct_additional']
        if 'extern_global_operations' in pop._specific_template.keys():
            extern_global_operations = pop._specific_template['extern_global_operations']
        if 'declare_spike_arrays' in pop._specific_template.keys():
            declare_spike = pop._specific_template['declare_spike_arrays']
        if 'declare_parameters_variables' in pop._specific_template.keys():
            declaration_parameters_variables = pop._specific_template['declare_parameters_variables']
        if 'declare_additional' in pop._specific_template.keys():
            declare_additional = pop._specific_template['declare_additional']
        if 'declare_FR' in pop._specific_template.keys():
            declare_FR = pop._specific_template['declare_FR']
        if 'declare_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            declare_delay = pop._specific_template['declare_delay']
        if 'access_parameters_variables' in pop._specific_template.keys():
            access_parameters_variables = pop._specific_template['access_parameters_variables']
        if 'access_additional' in pop._specific_template.keys():
            access_additional = pop._specific_template['access_additional']
        if 'init_parameters_variables' in pop._specific_template.keys():
            init_parameters_variables = pop._specific_template['init_parameters_variables']
        if 'init_spike' in pop._specific_template.keys():
            init_spike = pop._specific_template['init_spike']
        if 'init_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            init_delay = pop._specific_template['init_delay']
        if 'init_FR' in pop._specific_template.keys():
            init_FR = pop._specific_template['init_FR']
        if 'init_additional' in pop._specific_template.keys():
            init_additional = pop._specific_template['init_additional']
        if 'reset_spike' in pop._specific_template.keys():
            reset_spike = pop._specific_template['reset_spike']
        if 'reset_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            reset_delay = pop._specific_template['reset_delay']
        if 'reset_additional' in pop._specific_template.keys():
            reset_additional = pop._specific_template['reset_additional']
        if 'test_spike_cond' in pop._specific_template.keys():
            test_spike_cond = pop._specific_template['test_spike_cond']
        if 'update_rng' in pop._specific_template.keys():
            update_rng = pop._specific_template['update_rng']
        if 'update_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            update_delay = pop._specific_template['update_delay']
        if 'update_max_delay' in pop._specific_template.keys() and pop.max_delay > 1:
            update_max_delay = pop._specific_template['update_max_delay']
        if 'update_global_ops' in pop._specific_template.keys():
            update_global_ops = pop._specific_template['update_global_ops']

        # Fill the template
        code = self._templates['population_header'] % {
            # version tag
            'annarchy_version': ANNarchy.__release__,
            # fill code templates
            'float_prec': ConfigManager().get('precision', self._net_id),
            'id': pop.id,
            'name': pop.name,
            'size': pop.size,
            'include_additional': include_additional,
            'include_profile': include_profile,
            'struct_additional': struct_additional,
            'extern_global_operations': extern_global_operations,
            'declare_spike_arrays': declare_spike,
            'declare_parameters_variables': declaration_parameters_variables,
            'declare_additional': declare_additional,
            'declare_delay': declare_delay,
            'declare_FR': declare_FR,
            'declare_profile': declare_profile,
            'access_parameters_variables': access_parameters_variables,
            'access_additional': access_additional,
            'init_parameters_variables': init_parameters_variables,
            'init_spike': init_spike,
            'init_delay': init_delay,
            'init_FR': init_FR,
            'init_additional': init_additional,
            'init_profile': init_profile,
            'reset_spike': reset_spike,
            'reset_delay': reset_delay,
            'reset_additional': reset_additional,
            'update_variables': update_variables,
            'test_spike_cond': test_spike_cond,
            'update_rng': update_rng,
            'update_delay': update_delay,
            'update_max_delay': update_max_delay,
            'update_global_ops': update_global_ops,
            'stop_condition': stop_condition,
            'size_in_bytes': size_in_bytes,
            'clear_container': clear_container
        }

        # remove right-trailing spaces
        code = remove_trailing_spaces(code)

        # Store the complete header definition in a single file
        with open(annarchy_dir+'/generate/net'+str(self._net_id)+'/pop'+str(pop.id)+'.hpp', 'w') as ofile:
            ofile.write(code)

        # Basic informations common to all populations
        pop_desc = {
            'include': """#include "pop%(id)s.hpp"\n""" % {'id': pop.id},
            'extern': """extern PopStruct%(id)s *pop%(id)s;\n"""% {'id': pop.id},
            'instance': """PopStruct%(id)s *pop%(id)s;\n"""% {'id': pop.id},
            'init': """    pop%(id)s->init_population();\n""" % {'id': pop.id}
        }

        # Generate the calls to be made in the main ANNarchy.cpp
        if len(pop.neuron_type.description['variables']) > 0 or 'update_variables' in pop._specific_template.keys():
            if update_variables != "":
                pop_desc['update'] = """\tpop%(id)s->update();\n""" % {'id': pop.id}
                if pop.neuron_type.type == "spike":
                    pop_desc['update'] += """\tpop%(id)s->spike_gather();\n""" % {'id': pop.id}
            else:
                if "spike_gather_code" in pop._specific_template.keys():
                    if pop._specific_template["spike_gather_code"] != "":
                        pop_desc['update'] = """\tpop%(id)s->spike_gather();\n""" % {'id': pop.id}

        if len(pop.neuron_type.description['random_distributions']) > 0:
            pop_desc['rng_update'] = """\tpop%(id)s->update_rng();\n""" % {'id': pop.id}

        if pop.max_delay > 1:
            pop_desc['delay_update'] = """\tpop%(id)s->update_delay();\n""" % {'id': pop.id}

        if len(pop.global_operations) > 0:
            pop_desc['gops_update'] = """\tpop%(id)s->update_global_ops();\n""" % {'id': pop.id}

        return pop_desc

    ##################################################
    # Reset compute sums
    ##################################################
    def reset_computesum(self, pop):
        """
        For rate-coded neurons each step the weighted sum of inputs is computed. The implementation
        codes of the computes_rate kernel expect zeroed arrays. The same applies for the AccProjections
        used for the computation of the BOLD signal.

        Hint: this method is called directly by CodeGenerator.
        """
        code = ""

        # HD: use set to remove doublons
        for target in sorted(set(pop.targets)):
            code += self._templates['rate_psp']['reset'] % {
                'id': pop.id,
                'name': pop.name,
                'target': target,
                'float_prec': ConfigManager().get('precision', self._net_id)
            }

        return code

    ##################################################
    # Delays
    ##################################################
    def _delay_code(self, pop):
        """
        Generate code for delayed variables, comprising of initialization
        and update codes.

        Parameters:
            * population object

        Templates:
            attribute_delayed

        TODO:
            extract several templates, reorganize template dictionary
        """
        # Retrieve the template
        delay_tpl = self._templates['attribute_delayed']

        # Declaration
        declare_code = """
    // Delayed variables"""

        if pop.neuron_type.type == "rate":
            for var in pop.delayed_variables:
                attr = self._get_attr(pop, var)
                attr_dict = {'name': attr['name'], 'type': attr['ctype']}

                if attr['locality'] == "local":
                    declare_code += """
    std::deque< std::vector< %(type)s > > _delayed_%(name)s; """ % attr_dict
                else:
                    declare_code += """
    std::deque< %(type)s > _delayed_%(name)s; """ % attr_dict
        else:
            # Spiking networks should only exchange spikes
            declare_code += """
    // Delays for spike population
    std::deque< std::vector<int> > _delayed_spike;
"""
            for var in pop.delayed_variables:
                attr = self._get_attr(pop, var)
                attr_dict = {'name': attr['name'], 'type': attr['ctype']}

                if attr['locality'] == "local":
                    declare_code += """
    std::deque< std::vector< %(type)s > > _delayed_%(name)s; """ % attr_dict
                else:
                    declare_code += """
    std::deque< %(type)s > _delayed_%(name)s; """ % attr_dict

        # Initialization
        init_code = """
        // Delayed variables"""
        update_code = ""
        reset_code = ""
        resize_code = """
        if(value <= max_delay){ // nothing to do
            return;
        }
        max_delay = value;
        """
        for var in pop.delayed_variables:
            attr = self._get_attr(pop, var)
            init_code += delay_tpl[attr['locality']]['init'] % {'name': attr['name'], 'type': attr['ctype']}
            update_code += delay_tpl[attr['locality']]['update'] % {'name' : var}
            reset_code += delay_tpl[attr['locality']]['reset'] % {'id': pop.id, 'name' : var}
            resize_code += delay_tpl[attr['locality']]['resize'] % {'id': pop.id, 'name' : var, 'type': attr['ctype']}

        # Delaying spike events is done differently
        if pop.neuron_type.type == 'spike':
            init_code += """
        _delayed_spike = std::deque< std::vector<int> >(max_delay, std::vector<int>());"""

            update_code += """
            _delayed_spike.push_front(spiked);
            _delayed_spike.pop_back();
"""
            reset_code += """
        _delayed_spike.clear();
        _delayed_spike = std::deque< std::vector<int> >(max_delay, std::vector<int>());"""

            resize_code += """
        _delayed_spike.resize(max_delay, std::vector<int>());
"""

        update_code = """
        if ( _active ) {
%(code)s
        }""" % {'code': update_code}

        # if profiling enabled, annotate with profiling code
        if self._prof_gen:
            update_code = self._prof_gen.annotate_update_delay(pop, update_code)

        return declare_code, init_code, update_code, resize_code, reset_code

    ##################################################
    # Local functions
    ##################################################
    def _local_functions(self, pop):
        """
        Definition of user-defined local functions attached to
        a neuron. These functions will take place in the
        population header.
        """
        # Local functions
        if len(pop.neuron_type.description['functions']) == 0:
            return ""

        declaration = """
    // Local functions
"""
        for func in pop.neuron_type.description['functions']:
            declaration += ' '*4 + func['cpp'] + '\n'

        return declaration

    ##################################################
    # Stop condition
    ##################################################
    def _stop_condition(self, pop):
        """
        Simulation can either end after a fixed point in time or
        dependent on a population related condition. The code for
        this is generated here and added to the ANNarchy.cpp/.cu
        file.
        """
        if not pop.stop_condition: # no stop condition has been defined
            return ""

        # Special case for early return based on emitted events
        if pop.stop_condition.replace(" ", "") == "spiked:any":
            stop_code = """
    // Stop condition (any)
    bool stop_condition() {
        return !spiked.empty();
    } """
            return stop_code

        elif pop.stop_condition.replace(" ", "") == "spiked:all":
            stop_code = """
    // Stop condition (all)
    bool stop_condition() {
        return spiked.size() == size;
    } """
            return stop_code

        # Process the stop condition
        pop.neuron_type.description['stop_condition'] = {'eq': pop.stop_condition}
        from ANNarchy.parser.Extraction import extract_stop_condition
        extract_stop_condition(pop.neuron_type.description, pop.net_id)

        # Retrieve the code
        condition = pop.neuron_type.description['stop_condition']['cpp']% {
            'id': pop.id,
            'local_index': "[i]",
            'semiglobal_index': '',
            'global_index': ''}

        # Generate the function
        if pop.neuron_type.description['stop_condition']['type'] == 'any':
            stop_code = """
    // Stop condition (any)
    bool stop_condition(){
        for(int i=0; i<size; i++)
        {
            if(%(condition)s){
                return true;
            }
        }
        return false;
    }
    """ % {'condition': condition}
        else:
            stop_code = """
    // Stop condition (all)
    bool stop_condition(){
        for(int i=0; i<size; i++)
        {
            if(!(%(condition)s)){
                return false;
            }
        }
        return true;
    }
    """ % {'condition': condition}

        return stop_code


    ##################################################
    # Mean firing rate
    ##################################################
    def _init_fr(self, pop):
        "Declares arrays for computing the mean FR of a spiking neuron"
        declare_FR = ""; init_FR = ""; reset_FR = ""
        if pop.neuron_type.description['type'] == 'spike':
            declare_FR = """
    // Mean Firing rate
    std::vector< std::queue<long int> > _spike_history;
    long int _mean_fr_window;
    %(float_prec)s _mean_fr_rate;
    void compute_firing_rate(%(float_prec)s window){
        if(window>0.0){
            _mean_fr_window = int(window/dt);
            _mean_fr_rate = 1000./window;
            if (_spike_history.empty())
                _spike_history = std::vector< std::queue<long int> >(size, std::queue<long int>());
        }
    };""" % {'float_prec': ConfigManager().get('precision', self._net_id)}
            init_FR = """
        // Mean Firing Rate
        _spike_history = std::vector< std::queue<long int> >();
        _mean_fr_window = 0;
        _mean_fr_rate = 1.0;"""
            reset_FR = """
        // Mean Firing Rate
        for (auto it = _spike_history.begin(); it != _spike_history.end(); it++) {
            if (!it->empty()) {
                auto empty_queue = std::queue<long int>();
                it->swap(empty_queue);
            }
        }
"""

        return declare_FR, init_FR, reset_FR

    def _update_fr(self, pop):
        "Computes the average firing rate based on history"
        mean_FR_push = ""; mean_FR_update = ""
        if pop.neuron_type.description['type'] == 'spike':
            mean_FR_push = """
                    // Store the event for the mean firing rate
                    if (_mean_fr_window > 0)
                        _spike_history[i].push(t);
            """
            mean_FR_update = """if (_mean_fr_window > 0) {
                for (int i = 0; i < size; i++) {
                    while((_spike_history[i].size() != 0)&&(_spike_history[i].front() <= t - _mean_fr_window)){
                        _spike_history[i].pop(); // Suppress spikes outside the window
                    }
                    r[i] = _mean_fr_rate * %(float_prec)s(_spike_history[i].size());
                }
            } """ % {'float_prec': ConfigManager().get('precision', self._net_id)}

        return mean_FR_push, mean_FR_update

    ##################################################
    # Global operations
    ##################################################
    def _update_globalops(self, pop):
        """
        Update of global functions is a call of pre-implemented
        functions defined in GlobalOperationTemplate. In case of
        OpenMP this calls will take place in the population header.
        """
        if len(pop.global_operations) == 0:
            return ""

        code = ""
        for op in pop.global_operations:
            code += """
            _%(op)s_%(var)s = %(op)s_value(%(var)s.data(), size);
""" % {'op': op['function'], 'var': op['variable']}

        return """
    if ( _active ){
%(code)s
    }""" % {'code': code}

    def _update_random_distributions(self, pop):
        """
        Generate the C++ for drawing pseudo-random numbers in each step.
        The random variables are drawn sequentially from the same source.
        """
        if len(pop.neuron_type.description['random_distributions']) == 0:
            return ""

        rng_code = self._templates['rng']['update']

        local_code = ""
        global_code = ""
        rng_dist_code = ""
        for rd in pop.neuron_type.description['random_distributions']:
            rng_dist_code += self._templates['rng']['dist_decl'] % {
                'rd_name': rd['name'],
                'rd_init': rd['definition'] % {
                    'id': pop.id,
                    'float_prec': ConfigManager().get('precision', self._net_id),
                    'global_index': ''
                }
            }

            if rd['locality'] == 'local':
                local_code += self._templates['rng'][rd['locality']]['update'] % {'id': pop.id, 'rd_name': rd['name'], 'index': 0}
            else:
                global_code += self._templates['rng'][rd['locality']]['update'] % {'id': pop.id, 'rd_name': rd['name']}

        # Final code consists of local and global variables
        final_code = rng_code % {
            'rng_dist': rng_dist_code,
            'update_rng_local': local_code,
            'update_rng_global': global_code
        }

        # if profiling enabled, annotate with profiling code
        if self._prof_gen:
            final_code = self._prof_gen.annotate_update_rng(pop, final_code)

        return final_code

    ##################################################
    # Neural variables
    ##################################################
    def _update_rate_neuron(self, pop):
        """
        Generate the code template for neural update step, more precise updating of variables.
        The code comprise of two major parts: global and local update, second one parallelized
        with an openmp for construct, if number of threads is greater than one and the number
        of neurons exceed a minimum amount of neurons ( defined as Global.OMP_MIN_NB_NEURONS)
        """
        if "update_variables" in pop._specific_template.keys():
            final_eq = pop._specific_template["update_variables"]
            # if profiling enabled, annotate with profiling code
            if self._prof_gen:
                final_eq = self._prof_gen.annotate_update_neuron(pop, final_eq)
            return final_eq

        code = ""
        id_dict = {
            'id': pop.id,
            'local_index': "[i]",
            'semiglobal_index': '',
            'global_index': ''
        }

        # Random distributions
        deps =[]
        for rd in pop.neuron_type.description['random_distributions']:
            for dep in rd['dependencies']:
                deps += dep

        # Global variables
        eqs = generate_equation_code(pop.neuron_type.description, locality='global', padding=3)
        eqs = eqs % id_dict
        if eqs.strip() != "":
            code += """
            // Updating the global variables
%(eqs)s
""" % {'eqs': eqs}

        # Gather pre-loop declaration (dt/tau for ODEs)
        pre_code =""
        for var in pop.neuron_type.description['variables']:
            if 'pre_loop' in var.keys() and len(var['pre_loop']) > 0:
                pre_code += var['ctype'] + ' ' + var['pre_loop']['name'] + ' = ' + var['pre_loop']['value'] + ';\n'
        code = tabify(pre_code, 3) % {'id': pop.id, 'local_index': "[i]", 'semiglobal_index': '', 'global_index': ''} + code

        eqs = ""
        # sum() must generate _sum___all__[i] = _sum_exc[i] + sum_inh[i] + ... at the beginning
        if '__all__' in pop.neuron_type.description['targets']:
            eqs += " "*16 + "// Sum over all targets\n"
            eqs += " "*16 + "_sum___all__[i] = "
            for target in pop.targets:
                eqs += "_sum_" + target + '[i] + '
            eqs = eqs[:-2]
            eqs += ';\n\n'

        # Local variables, evaluated with SIMD if possible
        eqs = generate_equation_code(pop.neuron_type.description, 'local', padding=4)
        eqs = eqs % id_dict

        if eqs.strip() != "":
            code_dict = {
                'eqs': eqs,
                'id': pop.id,
                'omp_simd': "#pragma omp simd" if not ConfigManager().get('disable_SIMD_Eq', self._net_id) else ""
            }
            code += """
        if( _active ) {
        #ifdef _TRACE_SIMULATION_STEPS
            std::cout << "    PopStruct%(id)s::update()" << std::endl;
        #endif

            // Updating the local variables
            %(omp_simd)s
            for(int i = 0; i < size; i++){
%(eqs)s
            }
        } // active
""" % code_dict

        # if profiling enabled, annotate with profiling code
        if self._prof_gen:
            code = self._prof_gen.annotate_update_neuron(pop, code)

        return code

    def _update_spiking_neuron(self, pop):
        """
        Update the ODEs which describe the behavior of a neuron. Contrary to
        rate-coded neurons, there is a refractoriness possible, which disables
        the evolution of the ODE for a fixed period of time.
        """
        if "update_variables" in pop._specific_template.keys():
            final_eq = pop._specific_template["update_variables"]
            # if profiling enabled, annotate with profiling code
            if self._prof_gen:
                final_eq = self._prof_gen.annotate_update_neuron(pop, final_eq)
            return final_eq

        id_dict = {
            'id': pop.id,
            'local_index': "[i]",
            'semiglobal_index': '',
            'global_index': ''
        }

        # Global variables
        global_code = ""
        eqs = generate_equation_code(pop.neuron_type.description, locality='global', mask_variable=None, padding=3) % id_dict
        if eqs.strip() != "":
            global_code += """
            // Updating the global variables
%(eqs)s
""" % {'eqs': eqs}

        # Is there a refractory period?
        has_refractory = "in_ref" if (pop.neuron_type.refractory or pop.refractory) else None

        # Gather pre-loop declaration (dt/tau for ODEs)
        pre_code = ""
        for var in pop.neuron_type.description['variables']:
            if 'pre_loop' in var.keys() and len(var['pre_loop']) > 0:
                pre_code += var['ctype'] + ' ' + var['pre_loop']['name'] + ' = ' + var['pre_loop']['value'] + ';\n'
        if len(pre_code) > 0:
            pre_code = """
            // Updating the step sizes
""" + tabify(pre_code, 3)
            global_code = pre_code % id_dict + global_code

        # Local variables, evaluated in parallel
        local_code = generate_equation_code(pop.neuron_type.description, locality='local', mask_variable=has_refractory, padding=4) % id_dict
        
        # Decrement of refractoriness
        if has_refractory:
            local_code += tabify("""
// Decrement the refractory period
refractory_remaining[i] -= (1 - in_ref[i]);
""", 4)

        # Compute state of the refractory variable
        if has_refractory:
            comp_inref = """
            // compute which neurons are in refractory
            for(int i = 0; i < size; i++){
                in_ref[i] = (refractory_remaining[i] > 0) ? 0 : 1;
            }
            """

        else:
            comp_inref = ""

        # If axonal events are defined, clear the container
        if pop.neuron_type.axon_spike:
            global_code = "axonal.clear();\n"+ global_code

        # finish code
        final_eq = """
        if( _active ) {
%(comp_inref)s

%(global_code)s
            // Updating local variables
            %(omp_simd)s
            for(int i = 0; i < size; i++){
%(local_code)s
            }
        } // active
""" % {
    'omp_simd': "#pragma omp simd" if not ConfigManager().get('disable_SIMD_Eq', self._net_id) else "",
    'comp_inref': comp_inref,
    'local_code': local_code,
    'global_code': global_code
    }

        # if profiling enabled, annotate with profiling code
        if self._prof_gen:
            final_eq = self._prof_gen.annotate_update_neuron(pop, final_eq)

        return final_eq

    def _spike_gather(self, pop):
        """
        Generate the code for emitting spike events
        """
        if "spike_gather_code" in pop._specific_template.keys():
            final_spike_gather = pop._specific_template["spike_gather_code"]
            if self._prof_gen:
                final_spike_gather = self._prof_gen.annotate_spike_cond(pop, final_spike_gather)

            return final_spike_gather

        id_dict = {
            'id': pop.id,
            'local_index': "[i]",
            'semiglobal_index': '',
            'global_index': ''
        }

        # Is there an axonal spike condition?
        if pop.neuron_type.axon_spike:
            # get the conditions for axonal and neural spike event
            axon_cond = pop.neuron_type.description['axon_spike']['spike_cond'] % id_dict
            neur_cond = pop.neuron_type.description['spike']['spike_cond'] % id_dict

            # state changes if axonal spike occur
            axon_reset = ""
            for eq in pop.neuron_type.description['axon_spike']['spike_reset']:
                axon_reset += """
                    %(reset)s
            """ % { 'reset': eq['cpp'] % id_dict }

            # Simply extent the spiking vector, as the axonal spike
            # either manipulate neuron state nor consider refractoriness.
            # TODO: what happens for plastic networks? As it seems to be unclear,
            # after discussion with JB it is currently disabled (HD: 21.01.2019)
            axon_spike_code = """
                // Axon Spike Event, only if there was not already an event
                if( (%(axon_condition)s) && !(%(neur_condition)s) ) {
                    axonal.push_back(i);

                    %(axon_reset)s
                }
""" % {
    'axon_condition': axon_cond,
    'neur_condition': neur_cond,
    'axon_reset': axon_reset
}
        else:
            axon_spike_code = ""

        # Process the spike condition
        cond = pop.neuron_type.description['spike']['spike_cond'] % id_dict

        # Is there a refractory period?
        has_refractory = True if (pop.neuron_type.refractory or pop.refractory) else False

        # Decrement refractory if necessary
        if has_refractory:
            # By default, it is refractory, but users can specify another one
            refrac_var = "refractory[i]"
            if isinstance(pop.neuron_type.refractory, str):
                found = False
                for param in pop.neuron_type.description["parameters"] + pop.neuron_type.description["variables"]:
                    if param["name"] == pop.neuron_type.refractory:
                        if param['locality'] == 'local':
                            refrac_var = "int(" + pop.neuron_type.refractory + "[i]/dt)"
                        else:
                            refrac_var = "int(" + pop.neuron_type.refractory + "/dt)"
                        found = True
                        break
                if not found:
                    Messages._error("refractory = "+ pop.neuron_type.refractory + ": parameter or variable does not exist.")

            # set the refractory value
            refrac_inc = "refractory_remaining[i] = %(refrac_var)s;"%{'refrac_var': refrac_var}
        else:
            refrac_inc = ""

        # Mean Firing rate
        mean_FR_push, mean_FR_update = self._update_fr(pop)

        # Reset equations
        reset = ""
        for eq in pop.neuron_type.description['spike']['spike_reset']:
            reset += """
                    %(reset)s
""" % { 'reset': eq['cpp'] % id_dict }

        # Gather code
        refrac_check = tabify("""
    // check if neuron is in refractory
    if (in_ref[i]==0)
        continue;
""", 3)

        final_spike_gather = """
        if( _active ) {
            spiked.clear();

            for (int i = 0; i < size; i++) {
%(refrac_check)s

                // Spike emission
                if(%(condition)s){ // Condition is met
                    // Reset variables
%(reset)s
                    // Store the spike
                    spiked.push_back(i);
                    last_spike[i] = t;

                    // Refractory period
                    %(refrac_inc)s
                    %(mean_FR_push)s
                }
%(axon_spike_code)s
            }

            // Update mean firing rate
            %(mean_FR_update)s
        } // active
""" % {
    'condition' : cond,
    'refrac_check': refrac_check if has_refractory else "",
    'reset': reset,
    'refrac_inc': refrac_inc,
    'mean_FR_push': mean_FR_push,
    'mean_FR_update': mean_FR_update,
    'axon_spike_code': axon_spike_code
}

        # if profiling enabled, annotate with profiling code
        if self._prof_gen:
            final_spike_gather = self._prof_gen.annotate_spike_cond(pop, final_spike_gather)

        return final_spike_gather



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import time
import csv
import matplotlib.pylab as plt

from ANNarchy.intern.ConfigManagement import get_global_config, _update_global_config, _check_paradigm
from ANNarchy.intern import Messages

class Profiler :
    """
    The Profiler module should help to understand the performance of a simulation
    using the ANNarchy neural simulator.

    Therefore are functions to investigate memory consumption and timeline
    information provided.
    """
    _instance = None

    _color_code = {
        "default": "blue",
        "compile": "green",
        "simulate": "red",
        "instantiate": "orange",
        # will be ignored in image
        "cpp core": "black"
    }
    def __init__(self):
        """
        Constructor
        """
        pass 

    def __new__(cls):
        """
        First call construction of the NetworkManager. No additional arguments are required.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        
        return cls._instance

    def enable_profiling(self):
        """
        Initialize profiler instance and register it in Global.
        """
        # enable c++ profiling
        _update_global_config('profiling', True)

        # initialize measurement
        self._basetime = time.time()
        self._entries = []
        self._cpp_profiler = None       # set during Compiler._instantiate()
        self.add_entry( self._basetime, self._basetime, "initialized" )

    def disable_profiling(self):
        _update_global_config('profiling', False)
        self.clear()

    @property
    def enabled(self):
        return get_global_config('profiling')

    def add_entry( self, t_entry, t_escape, label, group="default" ):
        """
        Add a function to timeline.

        :param t_entry: entry time point of the function
        :param t_escape: escape time point of the function
        :param label: label of the function
        *:param group: which group does the function belong to (determines color code, default="default")
        """
        if group not in self._color_code.keys():
            # unknown group will be set to default values
            group = "default"

        self._entries.append( (t_entry, t_escape, label, group) )

    def update_entry( self, t_entry, t_escape, label, group ):
        """
        The profile entries are a list of tuples. Therefore such an entry can
        not modified easily.
        """
        if group not in self._color_code.keys():
            # unknown group will be set to default values
            group = "default"

        found = False
        for idx_t, (_,_,it_label, it_group) in enumerate(self._entries):
            if label == it_label and group == it_group:
                tmp = list(self._entries[idx_t])
                tmp[0] = t_entry
                tmp[1] = t_escape
                self._entries[idx_t] = tuple(tmp)
                found = True

        if not found:
            Messages._warning("Profiler.update_entry(): the entry was not found ...")

    def clear(self):
        """
        Clear all recorded time points.
        """
        self._entries.clear()

    def print_profile(self):
        """
        Print the content to console.
        """
        divided = ["cpp core", "instantiate", "compile"]

        for t_start, t_end, label, group in self._entries:
            if group not in divided: # Python functions
                print(label, ":", t_end-t_start, "seconds")

            if group == "compile":
                if label == "overall":
                    print("compile:", t_end-t_start, "seconds")
                else:
                    print("-", label, t_end-t_start, "seconds")

            if group == "instantiate":
                if label == "overall":
                    print("instantiate:", t_end-t_start, "seconds")
                else:
                    print("-", label, t_end-t_start, "seconds")

            if group == "cpp core": # CPP functions
                if t_start == 0.0:
                    continue

                if label == "overall":
                    print("-", label,":", t_start, "seconds (", t_end, "% )")
                else:
                    print("  -", label,":", t_start, "seconds (", t_end, "% )")

    def store_cpp_time_as_csv(self):
        """
        Store the measured timings on the C++ core as .csv to
        be further processed e. g. using pandas.
        """
        if _check_paradigm("cuda", 0): # TODO get net_id
            fname = "profile_cuda.csv"
        else:
            fname = "profile_omp_"+str(get_global_config('num_threads'))+"threads.csv"

        with open(get_global_config('profile_out')+'/'+fname, mode='w') as Datafile:
            csv_writer = csv.writer(Datafile, delimiter=',', quotechar=' ', quoting=csv.QUOTE_MINIMAL)

            for t_start, t_end, label, group in self._entries:
                # skip Python functions
                if group != "cpp core":
                    continue

                # non-defined function
                if t_start == 0.0:
                    continue

                # CPP functions
                if group == "cpp core":
                    csv_writer.writerow( (label, t_start, t_end, ) )

    def get_cpp_times(self):
        """
        Returns a dicitionary with all measured cpp-timings.
        """
        measurement = {}
        for t_start, t_end, label, group in self._entries:
            # skip Python functions
            if group != "cpp core":
                continue

            # non-defined function
            if t_start == 0.0:
                continue

            # CPP functions
            if group == "cpp core":
                measurement[label] = t_end
        return measurement

    def show_timeline(self, store_graph=False):
        """
        Visualize the timeline.
        """
        f, ax = plt.subplots()

        scale_param = 1.0 # origin data is in second

        for ts in self._entries:
            # Overview
            ax.barh(4, (ts[1]-ts[0])*scale_param, left=(ts[0]-self._basetime)*scale_param, color=self._color_code["default"])
        
            if ( ts[3] == "compile"):
                bar = ax.barh(3, (ts[1]-ts[0])*scale_param, left=(ts[0]-self._basetime)*scale_param, color=self._color_code[ts[3]], edgecolor="k", linewidth=2.0)
                patch = bar.patches[-1]
                
                bl = patch.get_xy()
                x = 0.5*patch.get_width() + bl[0]
                y = bl[1] - 0.1
                ax.text(x, y, ts[2], ha='center',va='center')

            if ( ts[3] == "simulate"):
                ax.barh(2, (ts[1]-ts[0])*scale_param, left=(ts[0]-self._basetime)*scale_param, color=self._color_code[ts[3]])

        ax.set_yticks([4,3,2])
        ax.set_yticklabels(["all", "compile()", "simulate()"])
        ax.set_xlabel("clock time [s]")
        
        if not store_graph:
            plt.show()
        else:
            f.savefig("timeline.png")


import ANNarchy
from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern.GlobalObjects import GlobalObjectManager
from ANNarchy.intern import Messages
from ANNarchy.core.Neuron import Neuron
from ANNarchy.core.Synapse import Synapse
from ANNarchy.core.PopulationView import PopulationView
import ANNarchy.parser.report.LatexParser as LatexParser
from ANNarchy.parser.AnalyseNeuron import analyse_neuron
from ANNarchy.parser.AnalyseSynapse import analyse_synapse
from ..Extraction import *

import numpy as np
import os

##################################
### Main method
##################################

def report_markdown(filename:str="./report.tex", standalone:bool=True, gather_subprojections:bool=False, title:str=None, author:str=None, date:str=None, net_id:int=0):
    """ Generates a .md file describing the network.

    *Parameters:*

    * **filename**: name of the .tex file where the report will be written (default: "./report.tex")
    * **standalone**: tells if the generated file should be directly compilable or only includable (ignored)
    * **gather_subprojections**: if a projection between two populations has been implemented as a multiple of projections between sub-populations, this flag allows to group them in the summary (default: False).
    * **net_id**: id of the network to be used for reporting (default: 0, everything that was declared)
    * **title**: title of the document (default: "Network description")
    * **author**: author of the document (default: "ANNarchy (Artificial Neural Networks architect)")
    * **date**: date of the document (default: empty)
    """

    # stdout
    Messages._print('Generating report in', filename)

    # Header
    if title == None:
        title = "Network description"
    if author == None:
        author = "ANNarchy (Artificial Neural Networks architect)"
    if date == None:
        date = ""
    header = """---
title: %(title)s
author: %(author)s
date: %(date)s
---
""" % {'title': title, 'author': author, 'date': date}

    # Structure
    structure = _generate_summary(net_id)

    # Neurons
    neuron_models = _generate_neuron_models(net_id)

    # Synapses
    synapse_models = _generate_synapse_models(net_id)

    # Parameters
    parameters = _generate_parameters(net_id, gather_subprojections)

    # Possibly create the directory if it does not exist
    path_name = os.path.dirname(filename)
    if not path_name in ["", "."]:
        if not os.path.exists(path_name):
            os.makedirs(path_name)
        
    with open(filename, 'w') as wfile:
        wfile.write(header)
        wfile.write(structure)
        wfile.write(neuron_models)
        wfile.write(synapse_models)
        wfile.write(parameters)


def _generate_summary(net_id):

    txt = """
# Structure of the network
"""

    # General information
    backend = 'default'
    if ConfigManager().get('paradigm', net_id) == 'cuda':
        backend = "CUDA"
    elif ConfigManager().get('paradigm', net_id) == "openmp" and ConfigManager().get('num_threads', net_id) > 1:
        backend = "OpenMP"
    txt +="""
* ANNarchy %(version)s using the %(backend)s backend.
* Numerical step size: %(dt)s ms.
""" % {'version': ANNarchy.__release__, 'backend': backend, 'dt': ConfigManager().get('dt', net_id)}

    # Populations
    if len(NetworkManager().get_network(net_id=net_id).get_populations()) > 0:
        headers = ["Population", "Size", "Neuron type"]
        populations = []
        for pop in NetworkManager().get_network(net_id=net_id).get_populations():
            # Find a name for the neuron
            neuron_name = "Neuron " + str(pop.neuron_type._rk_neurons_type) if pop.neuron_type.name in Neuron._default_names.values() \
                else pop.neuron_type.name

            populations.append([
                pop.name, 
                pop.geometry if len(pop.geometry)>1 else pop.size, 
                neuron_name])

        txt += """
## Populations

"""
        txt += _make_table(headers, populations)


    # Projections
    if len(NetworkManager().get_network(net_id=net_id).get_projections()) > 0 :
        headers = ["Source", "Destination", "Target", "Synapse type", "Pattern"]
        projections = []
        for proj in NetworkManager().get_network(net_id=net_id).get_projections():
            # Find a name for the synapse
            synapse_name = "Synapse " + str(proj.synapse_type._rk_synapses_type) if proj.synapse_type.name in Synapse._default_names.values() \
                else proj.synapse_type.name

            projections.append([
                proj.pre.name, 
                proj.post.name, 
                LatexParser._format_list(proj.target, ' / '),
                synapse_name,
                proj.connector_description
                ])

        txt += """
## Projections

"""
        txt += _make_table(headers, projections)

    # Monitors
    if len(NetworkManager().get_network(net_id=net_id).get_monitors()) > 0:
        headers = ["Object", "Variables", "Period"]
        monitors = []
        for monitor in NetworkManager().get_network(net_id=net_id).get_monitors():
            monitors.append([
                monitor.object.name + (" (subset)" if isinstance(monitor.object, PopulationView) else ""), 
                LatexParser._format_list(monitor.variables, ', '),
                monitor.period
                ])

        txt += """
## Monitors

"""
        txt += _make_table(headers, monitors)

    # Functions
    if GlobalObjectManager().number_functions() > 0 :
        txt += """## Functions

"""
        for _, func in GlobalObjectManager().get_functions():
            txt += LatexParser._process_functions(func, begin="$$", end="$$\n\n")

    return txt


# Neuron template
neuron_tpl = """
## %(name)s

%(description)s

**Parameters:**

%(parameters)s

**Equations:**

%(eqs)s
"""
def _generate_neuron_models(net_id):
    txt = """
# Neuron models
"""
    for idx, neuron in enumerate(GlobalObjectManager().get_neuron_types()):

        # Name
        if neuron.name in Neuron._default_names.values(): # name not set
            neuron_name = "Neuron " + str(neuron._rk_neurons_type)
        else:
            neuron_name = neuron.name

        # Description
        description = neuron.short_description
        if description == None:
            description = "Spiking neuron." if neuron.type == 'spike' else 'Rate-coded neuron'

        # Parameters
        parameters = extract_parameters(neuron.parameters, neuron.extra_values, object_type='neuron', net_id=net_id)
        parameters_list = [
            ["$" + LatexParser._latexify_name(param['name'], []) + "$", param['init'], 
                _adapt_locality_neuron(param['locality']), param['ctype']] 
                    for param in parameters]

        parameters_headers = ["Name", "Default value", "Locality", "Type"]
        parameters_table = _make_table(parameters_headers, parameters_list)

        if len(parameters) == 0:
            parameters_table = "$$\\varnothing$$"

        # Generate the code for the equations
        variables, spike_condition, spike_reset = LatexParser._process_neuron_equations(neuron, net_id)
        
        eqs = _process_variables(variables, neuron=True)

        # Spiking neurons
        if neuron.type == 'spike':

            reset_txt = "* Emit a spike a time $t$.\n"
            for r in spike_reset:
                reset_txt += "* $" + r + "$\n"

            eqs += """
**Spike emission:**

if $%(condition)s$ :

%(reset)s
""" % {'condition': spike_condition, 'reset': reset_txt}


        # Possible function
        if not neuron.functions == None:
            eqs += """
**Functions**

%(functions)s
""" % {'functions': LatexParser._process_functions(neuron.functions, begin="$$", end="$$\n\n")}

        # Finalize the template
        txt += neuron_tpl % {   'name': neuron_name, 
                                'description': description, 
                                'parameters': parameters_table,
                                'eqs': eqs}

    return txt

def _adapt_locality_neuron(l):
    d = {
        'local': "per neuron",
        'semiglobal': "per population",
        'global': "per population"
    }
    return d[l]

# Synapse template
synapse_tpl = """
## %(name)s

%(description)s

**Parameters:**

%(parameters)s

**Equations:**

%(eqs)s
%(psp)s
"""
def _generate_synapse_models(net_id):
    txt = """
# Synapse models
"""

    for idx, synapse in enumerate(GlobalObjectManager().get_synapse_types()):

        # Do not document default synapses
        if synapse.name == "-":
            continue

        # Find a name for the synapse
        synapse_name = "Synapse " + str(synapse._rk_synapses_type) if synapse.name in Synapse._default_names.values() else synapse.name

        # Description
        description = synapse.short_description
        if description == None:
            description = "Spiking synapse." if synapse.type == 'spike' else 'Rate-coded synapse'

        # Parameters
        parameters = extract_parameters(synapse.parameters, synapse.extra_values, object_type='synapse', net_id=net_id)
        parameters_list = [
            ["$" + LatexParser._latexify_name(param['name'], []) + "$", param['init'], _adapt_locality_synapse(param['locality']), param['ctype']] 
                for param in parameters]

        parameters_headers = ["Name", "Default value", "Locality", "Type"]
        parameters_table = _make_table(parameters_headers, parameters_list)

        if len(parameters) == 0:
            parameters_table = "$$\\varnothing$$"

        # Generate the code for the equations
        psp, variables, pre_desc, post_desc = LatexParser._process_synapse_equations(synapse, net_id=net_id)

        eqs = _process_variables(variables, neuron = False)

        # PSP
        if synapse.type == "rate":
            psp = """
**Weighted sum:**

$$%(transmission)s$$
"""  % {'transmission': psp}
        elif synapse.type == "spike" and psp != "":
            psp = """
**Continuous transmission:**

$$%(transmission)s$$
"""  % {'transmission': psp}
        else:
            psp = ""

        # Pre- and post-events
        if synapse.type == "spike":
            if len(pre_desc) > 0:
                eqs += """
**Pre-synaptic event at $t_\\text{pre} + d$:**
"""
                for pre in pre_desc:
                    eqs += "$$"+pre+"$$\n"
            if len(post_desc) > 0:
                eqs += """
**Post-synaptic event at $t_\\text{post}$:**
"""
                for post in post_desc:
                    eqs += "$$"+post+"$$\n"

        # Finalize the template
        txt += synapse_tpl % {  'name': synapse_name, 
                                'description': description, 
                                'psp': psp,
                                'parameters': parameters_table,
                                'eqs': eqs}

    return txt

def _adapt_locality_synapse(l):
    d = {
        'local': "per synapse",
        'semiglobal': "per post-synaptic neuron",
        'global': "per projection"
    }
    return d[l]


def _generate_parameters(net_id, gather_subprojections):
    txt = """
# Parameters
"""


    constants = NetworkManager().get_network(net_id).get_constants()

    # Constants
    if len(constants) > 0:
        txt += """
## Constants
"""
        constants_list = [
            ["$" + LatexParser._latexify_name(constant.name, []) + "$",  constant.value]
                for constant in constants]

        constants_headers = ["Name", "Value"]
        txt += _make_table(constants_headers, constants_list)

    # Population parameters
    txt += """
## Population parameters

"""
    parameters_list = []
    for rk, pop in enumerate(NetworkManager().get_network(net_id=net_id).get_populations()):    
        neuron_name = "Neuron " + str(pop.neuron_type._rk_neurons_type) if pop.neuron_type.name in Neuron._default_names.values() \
            else pop.neuron_type.name

        for idx, param in enumerate(pop.parameters):
            val = pop.init[param]
            if isinstance(val, (list, np.ndarray)):
                val = "$[" + str(np.array(val).min()) + ", " + str(np.array(val).max()) + "]$"
            parameters_list.append(
                [   LatexParser.pop_name(pop.name) if idx==0 else "", 
                    neuron_name if idx==0 else "", 
                    "$" + LatexParser._latexify_name(param, []) + "$", 
                    val ] )

    population_headers = ["Population", "Neuron type", "Name", "Value"]
    txt += _make_table(population_headers, parameters_list)

    # Projection parameters
    txt += """
## Projection parameters

"""
    if gather_subprojections:
        projections = []
        for proj in NetworkManager().get_network(net_id=net_id).get_projections():
            for existing_proj in projections:
                if proj.pre.name == existing_proj.pre.name and proj.post.name == existing_proj.post.name \
                    and proj.target == existing_proj.target : # TODO
                    break
            else:
                projections.append(proj)
    else:
        projections = NetworkManager().get_network(net_id=net_id).get_projections()

    parameters_list = []
    for rk, proj in enumerate(projections):
        for idx, param in enumerate(proj.parameters):
            if param == 'w':
                continue
            if idx == 0:
                proj_name = "%(pre)s  $\\rightarrow$ %(post)s with target %(target)s" % {
                    'pre': LatexParser.pop_name(proj.pre.name), 
                    'post': LatexParser.pop_name(proj.post.name), 
                    'target': LatexParser._format_list(proj.target, ' / ')}
            else:
                proj_name = ""
            
            synapse_name = "Synapse " + str(proj.synapse_type._rk_synapses_type) if proj.synapse_type.name in Synapse._default_names.values() \
                else proj.synapse_type.name
            
            val = proj.init[param]
            if isinstance(val, (list, np.ndarray)):
                val = "$[" + str(np.array(val).min()) + ", " + str(np.array(val).max()) + "]$"
            parameters_list.append(
                [   proj_name, 
                    synapse_name if idx == 0 else "",
                    "$" + LatexParser._latexify_name(param, []) + "$", 
                    val ] )

    projection_headers = ["Projection", "Synapse type", "Name", "Value"]
    txt += _make_table(projection_headers, parameters_list)

    return txt



def _process_variables(variables, neuron=True):
    eqs = ""
    for var in variables:
        # Min value
        if 'min' in var['bounds'].keys():
            min_val = ", minimum: " + str(var['bounds']['min'])
        else:
            min_val =""
        # Max value
        if 'max' in var['bounds'].keys():
            max_val = ", maximum: " + str(var['bounds']['max'])
        else:
            max_val =""
        # Method
        if var['ode']:
            method = ", " + find_method(var) + " numerical method"
        else:
            method = ""

        eqs += """
* Variable %(name)s : %(locality)s, initial value: %(init)s%(min)s%(max)s%(method)s

$$
%(code)s
$$
""" % { 'name': "$" + LatexParser._latexify_name(var['name'], []) + "$", 
        'code': var['latex'],
        'locality': _adapt_locality_neuron(var['locality']) if neuron else _adapt_locality_synapse(var['locality']),
        'init': var['init'],
        'min': min_val,
        'max': max_val,
        'method': method}

    return eqs

def _make_table(header, data):
    "Creates a markdown table from the data, with headers."

    nb_col = len(header)
    nb_data = len(data)

    # Compute the maximum size of each column
    max_size = [len(header[c]) + 4 for c in range(nb_col)]
    for c in range(nb_col):
        for e in range(nb_data):
            max_size[c] = max(max_size[c], len(str(data[e][c])))

    # Create the table
    table= "| "
    for c in range(nb_col):
        table += "**" + header[c] + "**" + " "*(max_size[c] - len(header[c]) - 4) + " | "
    table += "\n| "
    for c in range(nb_col):
        table += "-"*max_size[c] + " | "
    table += "\n"
    for e in range(nb_data):
        table += "| "
        for c in range(nb_col):
            table += str(data[e][c]) + " "*(max_size[c] - len(str(data[e][c]))) + " | "    
        table += "\n"    
    table += "\n"


    return table




"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern import Messages
from ANNarchy.core import Global

class DecodingProjection(SpecificProjection):
    """
    Decoding projection to transform spike trains into firing rates.

    The pre-synaptic population must be a spiking population, while the post-synaptic one must be rate-coded.

    Pre-synaptic spikes are accumulated for each post-synaptic neuron. A sliding window can be used to smoothen the results with the ``window`` parameter.

    The decoded firing rate is accessible in the post-synaptic neurons with `sum(target)`.

    The projection can be connected using any method available in `Projection` (although all-to-all or many-to-one makes mostly sense). Delays are ignored.

    The weight value allows to scale the firing rate: if you want a pre-synaptic firing rate of 100 Hz to correspond to a post-synaptic rate of 1.0, use ``w = 1./100.``.

    Example:

    ```python
    net = ann.Network()

    pop1 = net.create(ann.PoissonPopulation(1000, rates=100.))
    pop2 = net.create(1, ann.Neuron(equations="r=sum(exc)"))

    proj = net.connect(ann.DecodingProjection(pop1, pop2, 'exc', window=10.0))
    proj.all_to_all(1.0)
    ```

    :param pre: pre-synaptic population.
    :param post: post-synaptic population.
    :param target: type of the connection.
    :param window: duration of the time window to collect spikes (default: dt).
    :param name: optional name.
    """
    def __init__(
            self, 
            pre:"Population", 
            post:"Population", 
            target:str, 
            window:float=0.0, 
            name:str=None, 
            copied:bool=False, 
            net_id=0):

        # Instantiate the projection
        SpecificProjection.__init__(self, pre, post, target, None, name, copied, net_id)

        # Check populations
        if not self.pre.neuron_type.type == 'spike':
            Messages._error('The pre-synaptic population of a DecodingProjection must be spiking.')

        if not self.post.neuron_type.type == 'rate':
            Messages._error('The post-synaptic population of a DecodingProjection must be rate-coded.')

        # Process window argument
        if window == 0.0:
            window = ConfigManager().get('dt', self.net_id)
        self.window = window

        # Disable openMP post-synaptic matrix split
        self._no_split_matrix = True

        # Not on CUDA
        if _check_paradigm('cuda', self.net_id):
            Messages._error('DecodingProjections are not available on CUDA yet.')

    def _copy(self, pre, post, net_id=None):
        "Returns a copy of the population when creating networks. Internal use only."
        copied_proj = DecodingProjection(
            pre=pre, post=post, target=self.target, window=self.window, 
            name=self.name, copied=True,
            net_id = self.net_id if net_id is None else net_id,
        )
        copied_proj._no_split_matrix = True
        return copied_proj

    def _generate_st(self):
        # Generate the code
        self._specific_template['declare_additional'] = """
    // Window
    int window = %(window)s;
    std::deque< std::vector< %(float_prec)s > > rates_history ;
""" % { 'window': int(self.window/ConfigManager().get('dt', self.net_id)), 'float_prec': ConfigManager().get('precision', self.net_id) }

        self._specific_template['init_additional'] = """
        rates_history = std::deque< std::vector< %(float_prec)s > >(%(window)s, std::vector< %(float_prec)s >(%(post_size)s, 0.0));
""" % { 'window': int(self.window/ConfigManager().get('dt', self.net_id)),'post_size': self.post.size, 'float_prec': ConfigManager().get('precision', self.net_id) }

        self._specific_template['psp_code'] = """
        if (pop%(id_post)s->_active) {
            std::vector< std::pair<int, int> > inv_post;
            std::vector< %(float_prec)s > rates = std::vector< %(float_prec)s >(%(post_size)s, 0.0);
            // Iterate over all incoming spikes
            for(int _idx_j = 0; _idx_j < pop%(id_pre)s->spiked.size(); _idx_j++){
                rk_j = pop%(id_pre)s->spiked[_idx_j];
                inv_post = inv_pre_rank[rk_j];
                nb_post = inv_post.size();
                // Iterate over connected post neurons
                for(int _idx_i = 0; _idx_i < nb_post; _idx_i++){
                    // Retrieve the correct indices
                    i = inv_post[_idx_i].first;
                    j = inv_post[_idx_i].second;

                    // Increase the post-synaptic conductance
                    rates[post_rank[i]] +=  %(weight)s;
                }
            }

            rates_history.push_front(rates);
            rates_history.pop_back();
            for(int i=0; i<post_rank.size(); i++){
                sum = 0.0;
                for(int step=0; step<window; step++){
                    sum += rates_history[step][post_rank[i]];
                }
                pop%(id_post)s->_sum_%(target)s[post_rank[i]] += sum / %(float_prec)s(window) * 1000. / dt / %(float_prec)s(pre_rank[i].size());
            }
        } // active
""" % { 
        'id_proj': self.id, 
        'id_pre': self.pre.id, 
        'id_post': self.post.id, 
        'target': self.target,
        'post_size': self.post.size, 
        'float_prec': ConfigManager().get('precision', self.net_id),
        'weight': "w" if self._has_single_weight() else "w[i][j]"
      }

        self._specific_template['psp_prefix'] = """
        int nb_post, i, j, rk_j, rk_post, rk_pre;
        %(float_prec)s sum;
""" % { 'float_prec': ConfigManager().get('precision', self.net_id) }

    def _generate_omp(self):
        # Generate the code
        self._specific_template['declare_additional'] = """
    // Window
    int window = %(window)s;
    std::deque< std::vector< %(float_prec)s > > rates_history ;
""" % { 'window': int(self.window/ConfigManager().get('dt', self.net_id)), 'float_prec': ConfigManager().get('precision', self.net_id) }

        self._specific_template['init_additional'] = """
        rates_history = std::deque< std::vector< %(float_prec)s > >(%(window)s, std::vector< %(float_prec)s >(%(post_size)s, 0.0));
""" % { 'window': int(self.window/ConfigManager().get('dt', self.net_id)),'post_size': self.post.size, 'float_prec': ConfigManager().get('precision', self.net_id) }

        self._specific_template['psp_code'] = """
        #pragma omp single
        {
            if (pop%(id_post)s->_active) {
                std::vector< std::pair<int, int> > inv_post;
                std::vector< %(float_prec)s > rates = std::vector< %(float_prec)s >(%(post_size)s, 0.0);
                // Iterate over all incoming spikes
                for(int _idx_j = 0; _idx_j < pop%(id_pre)s->spiked.size(); _idx_j++){
                    rk_j = pop%(id_pre)s->spiked[_idx_j];
                    inv_post = inv_pre_rank[rk_j];
                    nb_post = inv_post.size();
                    // Iterate over connected post neurons
                    for(int _idx_i = 0; _idx_i < nb_post; _idx_i++){
                        // Retrieve the correct indices
                        i = inv_post[_idx_i].first;
                        j = inv_post[_idx_i].second;

                        // Increase the post-synaptic conductance
                        rates[post_rank[i]] +=  %(weight)s;
                    }
                }

                rates_history.push_front(rates);
                rates_history.pop_back();
                for(int i=0; i<post_rank.size(); i++){
                    sum = 0.0;
                    for(int step=0; step<window; step++){
                        sum += rates_history[step][post_rank[i]];
                    }
                    pop%(id_post)s->_sum_%(target)s[post_rank[i]] += sum / %(float_prec)s(window) * 1000. / dt / %(float_prec)s(pre_rank[i].size());
                }
            } // active
        }
""" %   { 
        'id_proj': self.id, 
        'id_pre': self.pre.id, 
        'id_post': self.post.id, 
        'target': self.target,
        'post_size': self.post.size, 
        'float_prec': ConfigManager().get('precision', self.net_id),
        'weight': "w" if self._has_single_weight() else "w[i][j]"
        }

        self._specific_template['psp_prefix'] = """
        int nb_post, i, j, rk_j, rk_post, rk_pre;
        %(float_prec)s sum;
""" % { 'float_prec': ConfigManager().get('precision', self.net_id) }

    def _generate_cuda(self):
        raise Global.ANNarchyException("The DecodingProjection is not available on CUDA devices.", True)


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

import numpy as np

from ANNarchy.intern.SpecificProjection import SpecificProjection
from ANNarchy.intern.ConfigManagement import _check_paradigm
from ANNarchy.intern import Messages
from ANNarchy.core import Global

class NormProjection(SpecificProjection):
    """
    Behaves like normal spike synapse but also generates a normalized conductance
    which means that the increase is normalized by the number of afferent synapses.

    Important is that the number of afferent synapses is across all dendrites
    of a neuron.

    :param pre: pre-synaptic Population or PopulationView.
    :param post: pre-synaptic Population or PopulationView.
    :param target: type of the connection.
    :param synapse: a ``Synapse`` instance.
    :param variable: target variable for normalized current.

    See also Projection.__init__().

    """
    def __init__(self, pre, post, target, variable, synapse=None, name=None, copied=False, net_id=0):
        # Instantiate the projection
        SpecificProjection.__init__(self, pre, post, target, synapse=synapse, name=name, copied=copied, net_id=net_id)
        self._variable = variable

        # Check populations
        if not self.pre.neuron_type.type == 'spike':
            Messages._error('The pre-synaptic population of an NormProjection must be spiking.')

        if not self.pre.neuron_type.type == 'spike':
            Messages._error('The post-synaptic population of an NormProjection must be spiking.')

        if synapse != None and not copied:
            Messages._error('NormProjection does not allow the usage of customized spiking synapses yet.')

        # Not on CUDA
        if _check_paradigm('cuda', self.net_id):
            Messages._error('NormProjections are not available on CUDA yet.')

        # Prevent automatic split of matrices
        self._no_split_matrix = True

    def _copy(self, pre, post, net_id=None):
        "Returns a copy of the population when creating networks. Internal use only."
        return NormProjection(pre=pre, post=post, target=self.target, variable=self._variable, synapse=self.synapse_type, name=self.name, copied=True, net_id=self.net_id if net_id is None else net_id)

    def _generate_st(self):
        """
        """
        # Sanity Check
        found = False
        for var in self.post.neuron_type.description['variables']:
            if var['name'] == self._variable:
                found = True
                break

        if not found:
            Messages._error("NormProjection: variable `"+self._variable+"` might be invalid. Please check the neuron model of population", self.post.name)

        # TODO: delays???
        if self.synapse_type.pre_axon_spike:
            pre_array = "tmp_spiked"
            pre_array_fusion = """
            std::vector<int> tmp_spiked = %(pre_array)s;
            tmp_spiked.insert( tmp_spiked.end(), pop%(id_pre)s->axonal.begin(), pop%(id_pre)s->axonal.end() );
""" % {'id_pre': self.pre.id, 'pre_array': 'pop'+str(self.pre.id)+'.spiked'}
        else:
            pre_array = 'pop'+str(self.pre.id)+'->spiked'
            pre_array_fusion = ""

        # nb_aff_synapse contains the number of all afferent synapses of this neuron
        # set after compile()
        if 'nb_aff_synapse' not in self.attributes:
            self.synapse_type.description['parameters'].append({'name': 'nb_aff_synapse',
                                                                'bounds': {},
                                                                'ctype': 'double',
                                                                'init': 1.0,
                                                                'locality': 'semiglobal'})
            self.attributes.append(['nb_aff_synapse'])

        # Get some more statements from user, normally done by the CodeGenerator
        if self._has_single_weight():
            psp_rside = "w"
        else:
            psp_rside = "w[i][j]" # default psp: g_target += w
        axon_code = ""
        indices = {
            'local_index': "[i][j]",
            'semiglobal_index': "[j]",
            'global_index': "[]"
        } # TODO: only true for openMP

        for var in self.synapse_type.description['pre_axon_spike']:
            if var['name'] == "g_target":
                psp_rside = var['cpp'].split("=")[1] % indices
            else:
                axon_code += var['cpp'] % indices

        # Only if needed. I don't really like the second loop, but it's for testing first
        if len(axon_code) > 0:
            axon_code = """
            for(int _idx_j = 0; _idx_j < pop%(id_pre)s->axonal.size(); _idx_j++){
                int rk_j = pop%(id_pre)s->axonal[_idx_j];
                auto inv_post_ptr = inv_pre_rank.find(rk_j);
                if (inv_post_ptr == inv_pre_rank.end())
                    continue;
                std::vector< std::pair<int, int> >& inv_post = inv_post_ptr->second;
                int nb_post = inv_post.size();

                // Iterate over connected post neurons
                for(int _idx_i = 0; _idx_i < nb_post; _idx_i++){
                    // Retrieve the correct indices
                    int i = inv_post[_idx_i].first;
                    int j = inv_post[_idx_i].second;

                    %(code)s
                }
            }
""" % {
    'id_pre': self.pre.id,
    'code': axon_code
}

        #
        # Generate Code Template Projection
        self._specific_template['psp_prefix'] = ""        
        self._specific_template['psp_code'] = """
        // Event-based summation
        if (_transmission && pop%(id_post)s->_active) {
            // Iterate over all incoming spikes
            %(spiked_array_fusion)s

            for(int _idx_j = 0; _idx_j < %(pre_array)s.size(); _idx_j++){
                int rk_j = %(pre_array)s[_idx_j];
                auto inv_post_ptr = inv_pre_rank.find(rk_j);
                if (inv_post_ptr == inv_pre_rank.end())
                    continue;
                std::vector< std::pair<int, int> >& inv_post = inv_post_ptr->second;
                int nb_post = inv_post.size();

                // Iterate over connected post neurons
                for(int _idx_i = 0; _idx_i < nb_post; _idx_i++){
                    // Retrieve the correct indices
                    int i = inv_post[_idx_i].first;
                    int j = inv_post[_idx_i].second;

                    pop%(id_post)s->g_%(target)s[post_rank[i]] += %(psp_rside)s;
                    pop%(id_post)s->%(var)s[post_rank[i]] += 1.0 / nb_aff_synapse[i];
                }
            }

            // axonal only
            %(axon_loop)s
        } // active
""" % { 'id_post': self.post.id,
        'target': self.target,
        'var': self._variable,
        'spiked_array_fusion': pre_array_fusion,
        'pre_array': pre_array,
        'psp_rside': psp_rside,
        'axon_loop': axon_code
    }

    def _generate_omp(self):
        """
        """
        # Sanity Check
        found = False
        for var in self.post.neuron_type.description['variables']:
            if var['name'] == self._variable:
                found = True
                break

        if not found:
            Messages._error("NormProjection: variable `"+self._variable+"` might be invalid. Please check the neuron model of population", self.post.name)

        # TODO: delays???
        if self.synapse_type.pre_axon_spike:
            pre_array = "tmp_spiked"
            pre_array_fusion = """
            std::vector<int> tmp_spiked = %(pre_array)s;
            tmp_spiked.insert( tmp_spiked.end(), pop%(id_pre)s->axonal.begin(), pop%(id_pre)s->axonal.end() );
""" % {'id_pre': self.pre.id, 'pre_array': 'pop'+str(self.pre.id)+'.spiked'}
        else:
            pre_array = 'pop'+str(self.pre.id)+'->spiked'
            pre_array_fusion = ""

        # nb_aff_synapse contains the number of all afferent synapses of this neuron
        # set after compile()
        if 'nb_aff_synapse' not in self.attributes:
            self.synapse_type.description['parameters'].append({'name': 'nb_aff_synapse',
                                                                'bounds': {},
                                                                'ctype': 'double',
                                                                'init': 1.0,
                                                                'locality': 'semiglobal'})
            self.attributes.append(['nb_aff_synapse'])

        # Get some more statements from user, normally done by the CodeGenerator
        if self._has_single_weight():
            psp_rside = "w"
        else:
            psp_rside = "w[i][j]" # default psp: g_target += w
        axon_code = ""
        indices = {
            'local_index': "[i][j]",
            'semiglobal_index': "[j]",
            'global_index': "[]"
        } # TODO: only true for openMP

        for var in self.synapse_type.description['pre_axon_spike']:
            if var['name'] == "g_target":
                psp_rside = var['cpp'].split("=")[1] % indices
            else:
                axon_code += var['cpp'] % indices

        # Only if needed. I don't really like the second loop, but it's for testing first
        if len(axon_code) > 0:
            axon_code = """
            for(int _idx_j = 0; _idx_j < pop%(id_pre)s->axonal.size(); _idx_j++){
                int rk_j = pop%(id_pre)s->axonal[_idx_j];
                auto inv_post_ptr = inv_pre_rank.find(rk_j);
                if (inv_post_ptr == inv_pre_rank.end())
                    continue;
                std::vector< std::pair<int, int> >& inv_post = inv_post_ptr->second;
                int nb_post = inv_post.size();

                // Iterate over connected post neurons
                for(int _idx_i = 0; _idx_i < nb_post; _idx_i++){
                    // Retrieve the correct indices
                    int i = inv_post[_idx_i].first;
                    int j = inv_post[_idx_i].second;

                    %(code)s
                }
            }
""" % {
    'id_pre': self.pre.id,
    'code': axon_code
}

        #
        # Generate Code Template Projection
        self._specific_template['psp_prefix'] = ""        
        self._specific_template['psp_code'] = """
        #pragma omp single
        {
            // Event-based summation
            if (_transmission && pop%(id_post)s->_active) {
                // Iterate over all incoming spikes
                %(spiked_array_fusion)s

                for(int _idx_j = 0; _idx_j < %(pre_array)s.size(); _idx_j++){
                    int rk_j = %(pre_array)s[_idx_j];
                    auto inv_post_ptr = inv_pre_rank.find(rk_j);
                    if (inv_post_ptr == inv_pre_rank.end())
                        continue;
                    std::vector< std::pair<int, int> >& inv_post = inv_post_ptr->second;
                    int nb_post = inv_post.size();

                    // Iterate over connected post neurons
                    for(int _idx_i = 0; _idx_i < nb_post; _idx_i++){
                        // Retrieve the correct indices
                        int i = inv_post[_idx_i].first;
                        int j = inv_post[_idx_i].second;

                        pop%(id_post)s->g_%(target)s[post_rank[i]] += %(psp_rside)s;
                        pop%(id_post)s->%(var)s[post_rank[i]] += 1.0 / nb_aff_synapse[i];
                    }
                }

                // axonal only
                %(axon_loop)s
            } // active
        }
""" % { 'id_post': self.post.id,
        'target': self.target,
        'var': self._variable,
        'spiked_array_fusion': pre_array_fusion,
        'pre_array': pre_array,
        'psp_rside': psp_rside,
        'axon_loop': axon_code
    }

    def _generate_cuda(self):
        raise NotImplementedError("BOLD monitor is not available for CUDA devices yet.")

def _update_num_aff_connections(net_id=0, verbose=False):
    """
    We need to set for all NormProjections the number of afferent
    connections for each projection (across multiple projections,
    otherwise pre_rank[i].size() would be sufficient.

    Attention:

    This function should be used only internally!!! The function is
    called during Compiler.compile() call.
    """
    # Do we need to execute this procedure?
    need_to_execute = False
    for proj in Global.projections():
        if isinstance(proj, NormProjection):
            need_to_execute = True
            break

    if need_to_execute == False:
        return

    # iterate over all populations
    for pop in Global.populations(net_id):
        nb_synapses = np.zeros((pop.size))

        # get afferent projections of this layer
        aff_proj = Global.projections(net_id, pop.name)

        # we accumulate the number of synapses per dendrite
        # across all afferent projections of this layer.
        for i, proj in enumerate(aff_proj):
            # nb synapses per dendrite is oriented at the post_ranks list. If a neuron
            # does not receive any connection in THIS projection there is no entry
            nb_synapses_per_dend = np.array(proj.nb_synapses_per_dendrite)
            if len(nb_synapses_per_dend) == 0:
                continue

            # Update global count
            for idx, rank in enumerate(proj.post_ranks):
                nb_synapses[rank] += nb_synapses_per_dend[idx]

        # set number of afferent connections for correct normalization
        for i, proj in enumerate(aff_proj):
            if not isinstance(proj, NormProjection):
                continue

            if verbose:
                print('Update:', proj.pre.name,'->', proj.post.name, '(', proj.target, ') -- (', i+1, 'of', len(aff_proj),')')
                if len(proj.post_ranks) != proj.post.size:
                    print('ranks:', proj.post_ranks)

            nb_synapses_per_dend = np.array(proj.nb_synapses_per_dendrite)

            if verbose:
                print('before:', nb_synapses_per_dend)
            for idx, rank in enumerate(proj.post_ranks):
                nb_synapses_per_dend[idx] = nb_synapses[rank]

            if verbose:
                print('after:', nb_synapses_per_dend)
            proj.cyInstance.set_semiglobal_attribute_all("nb_aff_synapse", nb_synapses_per_dend, "double")



"""
CMakeLists.txt templates.

:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Linux, Seq or OMP
linux_omp_template = """# CMakeLists.txt generated by ANNarchy
cmake_minimum_required(VERSION 3.16)

set(MODULE_NAME "ANNarchyCore%(net_id)s")
set(CMAKE_CXX_COMPILER "%(compiler)s")

project(${MODULE_NAME} LANGUAGES CXX)

# Find Python and add include paths
find_package(Python COMPONENTS Interpreter Development NumPy)
if (Python_FOUND)
include_directories(
    ${Python_INCLUDE_DIRS}
    ${Python_NumPy_INCLUDE_DIRS}
)
endif()

# Detect the installed nanobind package and import it into CMake
# see also: https://nanobind.readthedocs.io/en/latest/building.html
execute_process(
    COMMAND "${Python_EXECUTABLE}" -m nanobind --cmake_dir
    OUTPUT_STRIP_TRAILING_WHITESPACE OUTPUT_VARIABLE NB_DIR
)
list(APPEND CMAKE_PREFIX_PATH "${NB_DIR}")
find_package(nanobind CONFIG REQUIRED)

# Additional paths (ANNarchy-related)
include_directories(
    %(annarchy_include)s
    %(thirdparty_include)s
)

# Additional compiler flags (-fPIC will is added already)
add_compile_options(%(cpu_flags)s)
add_link_options(%(cpu_flags)s)

# Compile source files and generate shared library
nanobind_add_module(
    # Target name
    ${MODULE_NAME}
    NB_DOMAIN ${MODULE_NAME}
    # source files
    ${MODULE_NAME}.cpp
    ANNarchy.cpp
)

# Check if OpenMP is available and enable it.
find_package(OpenMP)
if(OpenMP_CXX_FOUND)
    target_link_libraries(${MODULE_NAME} PUBLIC OpenMP::OpenMP_CXX)
endif()

# Set the required C++ standard
target_compile_features(${MODULE_NAME} PUBLIC cxx_std_14)

# Nanobind generates by default ${MODULE_NAME}.cpython[version]-...
# we need to shorten that for easier handling.
set_target_properties(${MODULE_NAME} PROPERTIES SUFFIX ".so")

# After successful compilation move the shared library
add_custom_command(
    TARGET ${MODULE_NAME} POST_BUILD
    COMMAND ${CMAKE_COMMAND} -E copy ${MODULE_NAME}.so ../../
)
"""

# Linux, CUDA
linux_cuda_template = """# CMakeLists.txt generated by ANNarchy
cmake_minimum_required(VERSION 3.16)

set(MODULE_NAME "ANNarchyCore%(net_id)s")
set(CMAKE_CXX_COMPILER "%(compiler)s")

if (CMAKE_VERSION GREATER_EQUAL 3.18)
%(set_cuda_arch)s
endif()

project(${MODULE_NAME} LANGUAGES CXX CUDA)

# Find Python and add include paths
find_package(Python COMPONENTS Interpreter Development NumPy)
if (Python_FOUND)
include_directories(
    ${Python_INCLUDE_DIRS}
    ${Python_NumPy_INCLUDE_DIRS}
)
endif()

# Detect the installed nanobind package and import it into CMake
# see also: https://nanobind.readthedocs.io/en/latest/building.html
execute_process(
    COMMAND "${Python_EXECUTABLE}" -m nanobind --cmake_dir
    OUTPUT_STRIP_TRAILING_WHITESPACE OUTPUT_VARIABLE NB_DIR
)
list(APPEND CMAKE_PREFIX_PATH "${NB_DIR}")
find_package(nanobind CONFIG REQUIRED)

# Additional paths (ANNarchy-related)
include_directories(
    %(annarchy_include)s
    %(thirdparty_include)s
)

# For lower CMake versions, setting these flags leads to strange
# behavior (HD: 29th July 2024)
if (CMAKE_VERSION GREATER_EQUAL 3.17)
    # Additional compiler flags (-fPIC will is added already)
    add_compile_options(%(cpu_flags)s)
    add_link_options(%(cpu_flags)s)
endif()

# Compile source files and generate shared library
nanobind_add_module(
    # Target name
    ${MODULE_NAME}
    NB_DOMAIN ${MODULE_NAME}
    # source files (will trigger above command)
    ${MODULE_NAME}.cpp
    ANNarchy.cpp
    ANNarchyKernel.cu
)

# Set the required C++ standard
target_compile_features(${MODULE_NAME} PUBLIC cxx_std_14)

# Nanobind generates by default ${MODULE_NAME}.cpython[version]-...
# we need to shorten that for easier handling.
set_target_properties(${MODULE_NAME} PROPERTIES SUFFIX ".so")

# After successful compilation move the shared library
add_custom_command(
    TARGET ${MODULE_NAME} POST_BUILD
    COMMAND ${CMAKE_COMMAND} -E copy ${MODULE_NAME}.so ../../
)
"""

# OSX, with clang, Seq only
osx_clang_template = """# CMakeLists.txt generated by ANNarchy
cmake_minimum_required(VERSION 3.16)

set(MODULE_NAME "ANNarchyCore%(net_id)s")
set(CMAKE_CXX_COMPILER "%(compiler)s")

project(${MODULE_NAME} LANGUAGES CXX)

# Find Python and add include paths
find_package(Python COMPONENTS Interpreter Development NumPy)
if (Python_FOUND)
include_directories(
    ${Python_INCLUDE_DIRS}
    ${Python_NumPy_INCLUDE_DIRS}
)
endif()

# Detect the installed nanobind package and import it into CMake
# see also: https://nanobind.readthedocs.io/en/latest/building.html
execute_process(
    COMMAND "${Python_EXECUTABLE}" -m nanobind --cmake_dir
    OUTPUT_STRIP_TRAILING_WHITESPACE OUTPUT_VARIABLE NB_DIR
)
list(APPEND CMAKE_PREFIX_PATH "${NB_DIR}")
find_package(nanobind CONFIG REQUIRED)

# Additional paths (ANNarchy-related)
include_directories(
    %(annarchy_include)s
    %(thirdparty_include)s
)

# Additional compiler flags (-fPIC will is added already)
add_compile_options(%(cpu_flags)s)
add_link_options(%(cpu_flags)s)

# Compile source files and generate shared library
nanobind_add_module(
    # Target name
    ${MODULE_NAME}
    NB_DOMAIN ${MODULE_NAME}
    # source files
    ${MODULE_NAME}.cpp
    ANNarchy.cpp
)

# Check if OpenMP is available and enable it.
find_package(OpenMP)
if(OpenMP_CXX_FOUND)
    target_link_libraries(${MODULE_NAME} PUBLIC OpenMP::OpenMP_CXX)
endif()

# Set the required C++ standard
target_compile_features(${MODULE_NAME} PUBLIC cxx_std_14)

# Nanobind generates by default ${MODULE_NAME}.cpython[version]-...
# we need to shorten that for easier handling.
set_target_properties(${MODULE_NAME} PROPERTIES SUFFIX ".dylib")

# After successful compilation move the shared library
add_custom_command(
    TARGET ${MODULE_NAME} POST_BUILD
    COMMAND ${CMAKE_COMMAND} -E copy ${MODULE_NAME}.dylib ../../
)
"""

# OSX, with gcc, OpenMP
osx_gcc_template = """# CMakeLists.txt generated by ANNarchy
cmake_minimum_required(VERSION 3.16)

set(MODULE_NAME "ANNarchyCore%(net_id)s")
set(CMAKE_CXX_COMPILER "%(compiler)s")

project(${MODULE_NAME} LANGUAGES CXX)

# Find Python and add include paths
find_package(Python COMPONENTS Interpreter Development NumPy)
if (Python_FOUND)
include_directories(
    ${Python_INCLUDE_DIRS}
    ${Python_NumPy_INCLUDE_DIRS}
)
endif()

# Detect the installed nanobind package and import it into CMake
# see also: https://nanobind.readthedocs.io/en/latest/building.html
execute_process(
    COMMAND "${Python_EXECUTABLE}" -m nanobind --cmake_dir
    OUTPUT_STRIP_TRAILING_WHITESPACE OUTPUT_VARIABLE NB_DIR
)
list(APPEND CMAKE_PREFIX_PATH "${NB_DIR}")
find_package(nanobind CONFIG REQUIRED)

# Additional paths (ANNarchy-related)
include_directories(
    %(annarchy_include)s
    %(thirdparty_include)s
)

# Additional compiler flags (-fPIC will is added already)
add_compile_options(%(cpu_flags)s)
add_link_options(%(cpu_flags)s)

# Compile source files and generate shared library
nanobind_add_module(
    # Target name
    ${MODULE_NAME}
    NB_DOMAIN ${MODULE_NAME}
    # source files
    ${MODULE_NAME}.cpp
    ANNarchy.cpp
)

# Check if OpenMP is available and enable it.
find_package(OpenMP)
if(OpenMP_CXX_FOUND)
    target_link_libraries(${MODULE_NAME} PUBLIC OpenMP::OpenMP_CXX)
endif()

# Set the required C++ standard
target_compile_features(${MODULE_NAME} PUBLIC cxx_std_14)

# Nanobind generates by default ${MODULE_NAME}.cpython[version]-...
# we need to shorten that for easier handling.
set_target_properties(${MODULE_NAME} PROPERTIES SUFFIX ".dylib")

# After successful compilation move the shared library
add_custom_command(
    TARGET ${MODULE_NAME} POST_BUILD
    COMMAND ${CMAKE_COMMAND} -E copy ${MODULE_NAME}.dylib ../../
)
"""



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();       
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

#############################################
##  Synaptic delay
#############################################
delay = {
# A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
"""
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': {
        'declare': """
    std::vector<int> delay;
    int max_delay;

    std::vector<std::vector<int>> get_delay() { return get_matrix_variable_all<int>(delay); }
    void set_delay(std::vector<std::vector<int>> value) { update_matrix_variable_all<int>(delay, value); }
    std::vector<int> get_dendrite_delay(int lil_idx) { return get_matrix_variable_row<int>(delay, lil_idx); }
    int get_max_delay() { return max_delay; }
    void set_max_delay() { this->max_delay = max_delay; }
""",
        'init': """
    delay = init_matrix_variable<int>(1);
    update_matrix_variable_all<int>(delay, delays);

    max_delay = %(pre_prefix)smax_delay;
""",
        'reset': ""
    },
    # An individual value for each synapse and a
    # buffer for spike events
    'nonuniform_spiking': {
        'declare': """
    std::vector<int> delay;
    int max_delay;
    int idx_delay;
    std::vector< std::vector< std::vector< int > > > _delayed_spikes;
""",
        'init': """
    delay = init_matrix_variable<int>(1);
    update_matrix_variable_all<int>(delay, delays);

    idx_delay = 0;
    max_delay = %(pre_prefix)smax_delay;
""",
        'reset': """
        while(!_delayed_spikes.empty()) {
            auto elem = _delayed_spikes.back();
            elem.clear();
            _delayed_spikes.pop_back();
        }

        idx_delay = 0;
        max_delay =  %(pre_prefix)smax_delay ;
        _delayed_spikes = std::vector< std::vector< std::vector< int > > >(max_delay, std::vector< std::vector< int > >(post_rank.size(), std::vector< int >()) );        
""",
        'pyx_struct':
"""
        # Non-uniform delay
        vector[vector[int]] delay
        int max_delay
        void update_max_delay(int)
        void reset_ring_buffer()
"""
    }    
}

###############################################################
# Rate-coded continuous transmission
###############################################################
ell_summation_operation = {
    'sum' : """
%(pre_copy)s
const %(idx_type)s nonvalue_idx = std::numeric_limits<%(idx_type)s>::max();
%(size_type)s ell_row_off, j;

%(idx_type)s nb_post = static_cast<%(idx_type)s>(post_ranks_.size());
for (%(idx_type)s i = 0; i < nb_post; i++) {
    %(idx_type)s rk_post = post_ranks_[i]; // Get postsynaptic rank

    sum = 0.0;
    ell_row_off = i * maxnzr_;
    for(j = ell_row_off; j < ell_row_off+maxnzr_; j++) {
        %(idx_type)s rk_pre = col_idx_[j];
        if (rk_pre == nonvalue_idx)
            break;

        sum += %(psp)s ;
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum;
}"""
}

###############################################################
# Rate-coded synaptic plasticity
###############################################################
update_variables = {
    'local': """
const %(idx_type)s nonvalue_idx = std::numeric_limits<%(idx_type)s>::max();

// Check periodicity
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L) ){
    // Global variables
    %(global)s

    // Local variables
    for(%(size_type)s i = 0; i < post_ranks_.size(); i++){
        rk_post = post_ranks_[i]; // Get postsynaptic rank
        // Semi-global variables
        %(semiglobal)s
        // Local variables
        for(size_t j = i*maxnzr_; j < (i+1)*maxnzr_; j++) {
            rk_pre = col_idx_[j]; // Get presynaptic rank
            if (rk_pre == nonvalue_idx)
                break;
    %(local)s
        }
    }
}
"""
}

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,
    
    'rate_coded_sum': ell_summation_operation,
    'vectorized_default_psp': {},
    'update_variables': update_variables
}

conn_ids = {
    'local_index': '[j]',
    'semiglobal_index': '[i]',
    'global_index': '',
    'post_index': '[rk_post]',
    'pre_index': '[rk_pre]',
    'delay_u' : '[delay-1]', # uniform delay
    'delay_nu' : '[delay[j]-1]' # nonuniform delay
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

pooling_template_omp = {
    'include_additional': '#include <limits>',

    # Declare the connectivity matrix
    'declare_connectivity_matrix': """
    // connectivity data
    std::vector< std::vector<int> > pre_coords;
    """,

    # Accessors for the connectivity matrix
    'access_connectivity_matrix': """
    // Accessor to connectivity data
    std::vector<int> get_post_ranks() { 
        std::vector<int> v(pre_coords.size());
        std::iota (std::begin(v), std::end(v), 0);
        return v;
    }
    std::vector< std::vector<int> > get_pre_coords() { return pre_coords; }
    void set_pre_coords(std::vector< std::vector<int> > coords) { pre_coords = coords; }
""",

    # Variables for the psp code
    'psp_prefix': """
        int rk_pre;
        %(float_prec)s sum=%(sum_default)s;
    """,

    # Override the monitor to avoid recording the weights
    'monitor_class':"",
    'monitor_export': "",
    'monitor_wrapper': "",

    # Clear connectivity
    'clear': """
        // pre-coords sub-lists
        for (auto it = pre_coords.begin(); it != pre_coords.end(); it++) {
            it->clear();
            it->shrink_to_fit();
        }
        // pre-coords top-list
        pre_coords.clear();
        pre_coords.shrink_to_fit();
""",

    # No additional variables
    'declare_parameters_variables': "",
    'access_parameters_variables': "",
    'export_parameters_variables': "",

    # nanobind wrapper
    'wrapper': """
    // Pooling ProjStruct%(id_proj)s
    nanobind::class_<ProjStruct%(id_proj)s>(m, "proj%(id_proj)s_wrapper")
        // Constructor
        .def(nanobind::init<>())

        // Flags
        .def_rw("_transmission", &ProjStruct%(id_proj)s::_transmission)
        .def_rw("_axon_transmission", &ProjStruct%(id_proj)s::_axon_transmission)
        .def_rw("_update", &ProjStruct%(id_proj)s::_update)
        .def_rw("_plasticity", &ProjStruct%(id_proj)s::_plasticity)

        // Connectivity
        .def("post_rank", &ProjStruct%(id_proj)s::get_post_ranks)        
        .def_rw("pre_coords", &ProjStruct%(id_proj)s::pre_coords)

        // Other methods
        .def("clear", &ProjStruct%(id_proj)s::clear);
""",
    
}

pooling_template_cuda = {
    'include_additional': '#include <cfloat>',

    # Declare the connectivity matrix
    'declare_connectivity_matrix': """
    std::vector< std::vector<int> > pre_coords;
    int *gpu_pre_coords;
    """,

    # Accessors for the connectivity matrix
    'access_connectivity_matrix': """
    // Accessor to pre-synaptic coordinates (upper left corner)
    std::vector< std::vector<int> > get_pre_coords() { return pre_coords; }
    void set_pre_coords(std::vector< std::vector<int> > coords) {
        // host side
        pre_coords = coords;

        // Flattening coords
        auto num_coords = coords.size();
        auto coord_width = coords[0].size();
        auto flat_coords = std::vector<int>(num_coords*coord_width, 0);
        for (auto i = 0; i < num_coords; i++) {
            for( auto j = 0; j < coord_width; j++ ) {
                flat_coords[i*coord_width+j] = coords[i][j];
            }
        }

        cudaMalloc((void**)&gpu_pre_coords, flat_coords.size()*sizeof(int));
        cudaMemcpy( gpu_pre_coords, flat_coords.data(), flat_coords.size()*sizeof(int), cudaMemcpyHostToDevice);
        auto err = cudaGetLastError();
        if (err != cudaSuccess) {
            std::cerr << "Pooling: " << cudaGetErrorString(err) << std::endl;
        }
    }
""",
    'init_connectivity_matrix': "",

    # nanobind wrapper
    'wrapper': """
    // Pooling ProjStruct%(id_proj)s
    nanobind::class_<ProjStruct%(id_proj)s>(m, "proj%(id_proj)s_wrapper")
        // Constructor
        .def(nanobind::init<>())

        // Flags
        .def_rw("_transmission", &ProjStruct%(id_proj)s::_transmission)
        .def_rw("_axon_transmission", &ProjStruct%(id_proj)s::_axon_transmission)
        .def_rw("_update", &ProjStruct%(id_proj)s::_update)
        .def_rw("_plasticity", &ProjStruct%(id_proj)s::_plasticity)

        // Connectivity

        // Attributes
        .def("pre_coords", &ProjStruct%(id_proj)s::pre_coords)

        // Other methods
        .def("clear", &ProjStruct%(id_proj)s::clear);
""",


    # Memory transfer of variables
    'host_device_transfer': "",
    'device_host_transfer': "",

    # Override the monitor to avoid recording the weights
    'monitor_class':"",
    'monitor_export': "",
    'monitor_wrapper': "",

    # Memory management
    'clear': """
    // pre-coords sub-lists (host-side)
    for (auto it = pre_coords.begin(); it != pre_coords.end(); it++) {
        it->clear();
        it->shrink_to_fit();
    }
    // pre-coords top-list (host-side)
    pre_coords.clear();
    pre_coords.shrink_to_fit();

    // device side
    cudaFree(gpu_pre_coords);
"""
}

cuda_op_code = {
    'min': """if ( local_r < local_res ) local_res = local_r;""",
    'max': """if ( local_r > local_res ) local_res = local_r;""",
    'mean': "local_res += local_r;"
}

#
# For really small kernels it turns out to be beneficial
# to perform the operation with a single thread per block.
cuda_pooling_code_2d_small_extent = {
    'psp_body': """__global__ void cu_pooling_proj%(id_proj)s ( %(float_prec)s* __restrict__ psp, const int num_centers, const int* __restrict__ centers, const %(float_prec)s* __restrict__ %(pre_var)s) {
    int bIdx = blockIdx.x;
    int tid = threadIdx.x;

    // shared storage need to be initialized
    extern %(float_prec)s __shared__ sdata[];
    sdata[tid] = %(sum_default)s;
    __syncthreads();

    int coord_idx = floor(tid / static_cast<%(float_prec)s>(%(col_extent)s)) + blockIdx.x * floor( blockDim.x / static_cast<%(float_prec)s>(%(col_extent)s) );
    if (coord_idx < num_centers) {
        int x_coords = centers[2*coord_idx];
        int y_coords = centers[2*coord_idx+1];

        int idx_x, idx_y;
        %(float_prec)s local_r = 0.0;
        %(float_prec)s local_res = %(sum_default)s;

        for( int x = 0; x < %(row_extent)s; x++ ) {
            idx_x = x_coords + x;
            if ( idx_x >= %(row_size)s )
                continue;

            for( int y = 0; y < %(col_extent)s; y++ ) {
                idx_y = y_coords + y;
                if ( idx_y >= %(col_size)s )
                    continue;

                local_%(pre_var)s = %(pre_var)s[ idx_x * %(col_size)s + idx_y ];
%(operation)s
            }
        }

        // store intermediate result
        sdata[tid] = local_res;
        __syncthreads();

        // Reduction in shared memory
%(operation_reduce)s
    }
};
""",
    'psp_invoke': """
void pooling_proj%(id_proj)s (RunConfig cfg, %(float_prec)s* psp, const int num_centers, const int* centers, const %(float_prec)s* %(pre_var)s) {
    cu_pooling_proj%(id_proj)s<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(
        psp, num_centers, centers, %(pre_var_invoke)s
    );
}
""",
    'psp_header': """void pooling_proj%(id_proj)s (RunConfig cfg, %(float_prec)s* psp, const int num_centers, const int* centers, const %(float_prec)s* %(pre_var)s);
""",
    'psp_call': """
    int coords_per_block = floor(32.0 / static_cast<%(float_prec)s>(%(col_extent)s));
    int num_blocks = ceil(static_cast<%(float_prec)s>(%(size_post)s) / static_cast<%(float_prec)s>(coords_per_block));
    int thread_per_block = %(col_extent)s * coords_per_block;
    int shared_mem_size = thread_per_block * sizeof(%(float_prec)s);
    pooling_proj%(id_proj)s(RunConfig(num_blocks, thread_per_block, shared_mem_size, proj%(id_proj)s.stream), pop%(id_post)s->gpu__sum_%(target)s, %(size_post)s, proj%(id_proj)s.gpu_pre_coords, pop%(id_pre)s->gpu_%(pre_var)s );
""",
    # The reduction stage is responsible to fuse the several local results within
    # the warp to the final result. ATTENTION: there are several results in this warp
    'reduce_code': {
        'max': """
    if ( tid %% %(col_extent)s == 0 ) {
        for(int y = 1; y < %(col_extent)s; y++)
            sdata[tid] = max(sdata[tid], sdata[tid+y]);

        psp[coord_idx] = sdata[tid];
    }
""",
    }
}

#
# Pooling implementation where a warp handles a row
# at once.
cuda_pooling_code_2d = {
    'psp_body': """__global__ void cu_pooling_proj%(id_proj)s ( %(float_prec)s* __restrict__ psp, const int shared_size, const int* __restrict__ centers, const %(float_prec)s* __restrict__ %(pre_var)s) {
    int bIdx = blockIdx.x;
    int tid = threadIdx.x;

    // get pre-synaptic coordinates
    int x_coords = centers[2*bIdx];
    int y_coords = centers[2*bIdx+1];

    // shared storage need to be initialized
    extern %(float_prec)s __shared__ sdata[];
    for(int x = tid; x < shared_size; x++)
        sdata[x] = %(sum_default)s;
    __syncthreads();

    // local variables
    int idx_x, idx_y;
    %(float_prec)s local_r = 0.0;
    %(float_prec)s local_res = %(sum_default)s;

    if (tid < %(col_extent)s) {

        // row-wise scan
        for( int x = 0; x < %(row_extent)s; x++ ) {
            idx_x = x_coords + x;
            if ( idx_x >= %(row_size)s )
                continue;

            for( int y = tid; y < %(col_extent)s; y += blockDim.x ) {
                idx_y = y_coords + y;
                if ( idx_y >= %(col_size)s )
                    continue;

                local_%(pre_var)s = %(pre_var)s[ idx_x * %(col_size)s + idx_y ];
%(operation)s
            }
        }

        // store intermediate result
        sdata[tid] = local_res;
        __syncthreads();

        // Reduction in shared memory
%(operation_reduce)s
    }
}
""",
    'psp_invoke': """
void pooling_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int shared_size, const int* centers, const %(float_prec)s* %(pre_var)s) {
    cu_pooling_proj%(id_proj)s<<< cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream >>>(
        psp, shared_size, centers, r
    );
}
""",
    'psp_header': """void pooling_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int shared_size, const int* centers, const %(float_prec)s* %(pre_var)s);
""",
    # Technically, we could use more than warp-size threads.
    # But as we often have small extents it is not required (HD: 27. Nov. 2020)
    'psp_call': """
    auto tpb = 32;
    auto shared_size = min(32, tpb);
    auto smem_size = 2 * shared_size * sizeof(%(float_prec)s);
    pooling_proj%(id_proj)s(RunConfig(%(size_post)s, tpb, smem_size, proj%(id_projs).stream), pop%(id_post)s->gpu__sum_%(target)s, 2*shared_size, proj%(id_proj)s.gpu_pre_coords, pop%(id_pre)s->gpu_%(pre_var)s );
""",
    # The reduction stage is responsible to fuse the
    # several local results within the warp to the final result
    'reduce_code': {
        'min': """if (tid < 16)
{
    volatile %(float_prec)s* smem = sdata;

    smem[tid] = min(smem[tid], smem[tid + 16]);
    smem[tid] = min(smem[tid], smem[tid + 8]);
    smem[tid] = min(smem[tid], smem[tid + 4]);
    smem[tid] = min(smem[tid], smem[tid + 2]);
    smem[tid] = min(smem[tid], smem[tid + 1]);
}
__syncthreads();

// write result for this block to global mem
if (tid == 0)
{
    psp[bIdx] = sdata[0];
}
""",
        'max': """if (tid < 16)
{
    volatile %(float_prec)s* smem = sdata;

    smem[tid] = max(smem[tid], smem[tid + 16]);
    smem[tid] = max(smem[tid], smem[tid + 8]);
    smem[tid] = max(smem[tid], smem[tid + 4]);
    smem[tid] = max(smem[tid], smem[tid + 2]);
    smem[tid] = max(smem[tid], smem[tid + 1]);
}
__syncthreads();

// write result for this block to global mem
if (tid == 0)
{
    psp[bIdx] = sdata[0];
}
""",
        'mean': """if (tid < 16)
{
    volatile %(float_prec)s* smem = sdata;

    smem[tid] = smem[tid] + smem[tid + 16];
    smem[tid] = smem[tid] + smem[tid + 8];
    smem[tid] = smem[tid] + smem[tid + 4];
    smem[tid] = smem[tid] + smem[tid + 2];
    smem[tid] = smem[tid] + smem[tid + 1];
}
__syncthreads();

// write result for this block to global mem
if (tid == 0)
{
    psp[bIdx] = sdata[0] / static_cast<%(float_prec)s>(%(row_extent)s * %(col_extent)s);
}
"""
    }
}

cuda_pooling_code_3d = {
    'psp_body': """__global__ void pooling_proj%(id_proj)s ( %(float_prec)s* __restrict__ psp, const int* __restrict__ centers, const %(float_prec)s* __restrict__ %(pre_var)s) {
    int bIdx = blockIdx.x;

    int x_coords = centers[3*bIdx];
    int y_coords = centers[3*bIdx+1];
    int z_coords = centers[3*bIdx+2];

    int idx_x, idx_y, idx_z;
    %(float_prec)s local_r = 0.0;
    %(float_prec)s local_res = %(sum_default)s;

    for (int x = 0; x < %(row_extent)s; x++ ) {
        idx_x = x_coords + x;
        if ( idx_x >= %(row_size)s )
            continue;

        for (int y = 0; y < %(col_extent)s; y++ ) {
            idx_y = y_coords + y;
            if ( idx_y >= %(col_size)s )
                continue;

            for (int z = 0; z < %(plane_extent)s; z++) {
                idx_z = z_coords + z;
                if ( idx_z >= %(plane_size)s )
                    continue;

                local_r = %(pre_var)s[ %(plane_size)s * (%(col_size)s * idx_x + idx_y) + idx_z ];
%(operation)s
            }
        }
    }

    %(final_result)s
}
""",
    'psp_invoke': """void pooling_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int* centers, const %(float_prec)s* %(pre_var)s) {
    pooling_proj%(id_proj)s<<< %(size_post)s, 1 >>>(psp, centers, %(pre_var)s);
}
""",
    'psp_header': """void pooling_proj%(id_proj)s(RunConfig cfg, %(float_prec)s* psp, const int* centers, const %(float_prec)s* %(pre_var)s);
""",
    'psp_call': """
    if (proj%(id_proj)s._transmission && pop%(id_post)s->_active ) {
        pooling_proj%(id_proj)s(RunConfig(%(size_post)s, 1, 0, proj%(id_proj)s.stream), pop%(id_post)s->gpu__sum_%(target)s, proj%(id_proj)s.gpu_pre_coords, pop%(id_pre)s->gpu_%(pre_var)s );

    #ifdef _DEBUG
        auto proj%(id_proj)s_pool_err = cudaDeviceSynchronize();
        if ( proj%(id_proj)s_pool_err != cudaSuccess) {
            std::cout << "Pooling projection %(id_proj)s - psp: " << cudaGetErrorString( proj%(id_proj)s_pool_err ) << std::endl;
        }
    #endif
    }
"""
}



"""
The SingleThread package does contain all code templates required for the single-thread
code generation in ANNarchy.

BaseTemplates:

    defines the basic defintions common to all sparse matrix formates, e. g. projection header

[FORMAT]_SingleThread:

    defines the format specific defintions for the currently available formats:

        * LIL: list-in-list
        * COO: coordinate
        * DIA: diagonal format
        * BSR: blocked compressed row
        * CSR: compressed sparse row
        * ELL: ELLPACK/ITPACK
        * ELLR: ELLPACK with row-length array
        * SELLR: sliced ELLPACK
        * HYB: hybrid format comprising of ELLPACK and coordinate
        * Dense: a full matrix representation

    there are some special purpose implementations:

        * _T suffix: a transposed implementation
        * _PV suffix: a specialized implementation for PopulationViews

"""
from . import LIL as LIL_SingleThread
from . import COO as COO_SingleThread
from . import DIA as DIA_SingleThread
from . import BSR as BSR_SingleThread
from . import CSR as CSR_SingleThread
from . import CSR_T as CSR_T_SingleThread
from . import ELL as ELL_SingleThread
from . import ELLR as ELLR_SingleThread
from . import SELL as SELL_SingleThread
from . import HYB as HYB_SingleThread
from . import Dense as Dense_SingleThread
from . import Dense_T as Dense_T_SingleThread
from . import Dense_PV as Dense_PV_SingleThread

__all__ = [
    "BaseTemplates",
    "LIL_SingleThread",
    "COO_SingleThread",
    "DIA_SingleThread",
    "BSR_SingleThread",
    "CSR_SingleThread", "CSR_T_SingleThread",
    "ELL_SingleThread", "ELLR_SingleThread", "SELL_SingleThread",
    "HYB_SingleThread",
    "Dense_SingleThread", "Dense_PV_SingleThread", "Dense_T_SingleThread"
]



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Authorized keywork for attributes
authorized_keywords = [
    # Init
    'init',
    # Bounds
    'min',
    'max',
    # Locality
    'population',
    'postsynaptic',
    'projection',
    # Numerical methods
    'explicit',
    'implicit',
    'semiimplicit',
    'exponential',
    'midpoint',
    'rk4',
    'runge-kutta4', # backward compatibility
    'exact',
    'event-driven',
    # Refractory
    'unless_refractory',
    # Type
    'int',
    'bool',
    'float',
    # Event-based
    'unless_post',
]

####################################
# Functions for string manipulation
####################################
import re
from ANNarchy.intern import Messages

def split_equation(definition):
    " Splits a description into equation and flags."
    try:
        equation, constraint = definition.rsplit(':', 1)
    except ValueError:
        equation = definition.strip() # there are no constraints
        constraint = None
    else:
        has_constraint = False
        for keyword in authorized_keywords:
            if keyword in constraint:
                has_constraint = True
        if has_constraint:
            equation = equation.strip()
            constraint = constraint.strip()
        else:
            equation = definition.strip() # there are no constraints
            constraint = None
    finally:
        return equation, constraint

def prepare_string(stream):
    """ 
    Splits up a multiline equation, remove comments and unneeded spaces or tabs.
    """
    expr_set = []
    # replace ; with new line and split the result up
    tmp_set = stream.replace(';', '\n').split('\n')
    for expr in tmp_set:
        expr = re.sub(r'\#[\s\S]+', ' ', expr)   # remove comments
        expr = re.sub(r'\s+', ' ', expr)     # remove additional tabs etc.
        if expr.strip() == '' or len(expr)==0: # through beginning line breaks or something similar empty strings are contained in the set
            continue
        expr_set.append(''.join(expr))
    return expr_set

def extract_name(equation, left=False):
    " Extracts the name of a parameter/variable by looking the left term of an equation."
    equation = equation.replace(' ','')
    if not left: # there is potentially an equal sign
        try:
            name = equation.split('=')[0]
        except: # No equal sign. Eg: baseline : init=0.0
            return equation.strip()

        # Search for increments
        operators = ['+=', '-=', '*=', '/=', '>=', '<=']
        for op in operators:
            if op in equation:
                return equation.split(op)[0]
    else:
        name = equation.strip()
        # Search for increments
        operators = ['+', '-', '*', '/']
        for op in operators:
            if equation.endswith(op):
                return equation.split(op)[0]
    # Check for error
    if name.strip() == "":
        Messages._error('the variable name can not be extracted from ' + equation)

    # Search for any operation in the left side
    operators = ['+', '-', '*', '/']
    ode = False
    for op in operators:
        if not name.find(op) == -1:
            ode = True
    if not ode: # variable name is alone on the left side
        return name
    # ODE: the variable name is between d and /dt
    name = re.findall(r"d([\w]+)/dt", name)
    if len(name) == 1:
        return name[0].strip()
    else:
        return '_undefined'


def extract_flags(constraint):
    """
    Extracts from all attributes given after : which are bounds (eg min=0.0 or init=0.1) and which are flags (eg postsynaptic, implicit...).
    """

    bounds = {}
    flags = []

    # Check if there are constraints at all
    if not constraint:
        return bounds, flags
    
    # Regexp explanation:
    # (?:[^,()]+|\((?:[^()]*)\))+  # Matches either:
    #   - Text not containing commas or parentheses
    #   - Parenthesized content (balanced)
    pattern = r'(?:[^,()]+|\([^()]*\))+' 
    
    for con in [param.strip() for param in re.findall(pattern, constraint)]:
        try: # bound of the form key = val
            key, value = con.split('=')
            bounds[key.strip()] = value.strip()
        except ValueError: # No equal sign = flag
            flags.append(con.strip())

    return bounds, flags


def process_equations(equations):
    """ 
    Takes a multi-string describing equations and returns a list of dictionaries, where:

    * 'name' is the name of the variable
    * 'eq' is the equation
    * 'constraints' is all the constraints given after the last :. _extract_flags() should be called on it.
    """
    def is_constraint(eq):
        " Internal method to determine if a string contains reserved keywords."
        eq = ',' +  eq.replace(' ', '') + ','
        for key in authorized_keywords:
            pattern = '([,]+)' + key + '([=,]+)'
            if re.match(pattern, eq):
                return True
        return False

    # All equations will be stored there, in the order of their definition
    variables = []
    try:
        equations = equations.replace(';', '\n').split('\n')
    except: # equations is empty
        return variables


    # Iterate over all lines
    for line in equations:
        # Skip empty lines
        definition = line.strip()
        if definition == '':
            continue

        # Remove comments
        com = definition.split('#')
        if len(com) > 1:
            definition = com[0]
            if definition.strip() == '':
                continue

        # Process the line
        try:
            equation, constraint = definition.rsplit(':', 1)
        
        except ValueError: # There is no :, only equation is concerned
            equation = definition
            constraint = ''
        
        else:   # there is a :
            # Check if the constraint contains the reserved keywords
            has_constraint = is_constraint(constraint)
        
            # If the right part of : is a constraint, just store it
            # Otherwise, it is an if-then-else statement
            if has_constraint:
                equation = equation.strip()
                constraint = constraint.strip()
            else:
                equation = definition.strip() # there are no constraints
                constraint = ''

        # Split the equation around operators = += -= *= /=, but not ==
        split_operators = re.findall(r'([\s\w\+\-\*\/\)]+)=([^=])', equation)
        if len(split_operators) == 1: # definition of a new variable
            # Retrieve the name
            eq = split_operators[0][0]
            if eq.strip() == "":
                Messages._print(equation)
                Messages._error('The equation can not be analysed, check the syntax.')

            name = extract_name(eq, left=True)
            if name in ['_undefined', '']:
                Messages._error('No variable name can be found in ' + equation)

            # Append the result
            variables.append({'name': name, 'eq': equation.strip(), 'constraint': constraint.strip()})
        elif len(split_operators) == 0:
            # Continuation of the equation on a new line: append the equation to the previous variable
            variables[-1]['eq'] += ' ' + equation.strip()
            variables[-1]['constraint'] += constraint
        else:
            Messages._error('Only one assignement operator is allowed per equation.')

    return variables



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

# Code which should be added prior to kernels
# (directly imported by CodeGenerator if needed)
additional_global_functions = ""

launch_config = {
    'init': """
        _nb_blocks = std::min<unsigned int>(block_row_size(), 65535);

        // must be multiple of tile_size^2
        unsigned int tile_size2 = get_tile_size() * get_tile_size();
        _threads_per_block = (32/tile_size2)*tile_size2;
        if (_threads_per_block == 0) {
            _threads_per_block = tile_size2;
        }

    #ifdef _DEBUG
        std::cout << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif

""",
    'update': """
        std::cout << "Adjustment of launch configuration for BSR is not supported yet." << std::endl;

    #ifdef _DEBUG
        std::cout << _nb_blocks << ", " << _threads_per_block << std::endl;
    #endif
"""
}

attribute_decl = {
    'local': """
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'semiglobal': """
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
""",
    'global': """
    // Global %(attr_type)s %(name)s
    %(type)s %(name)s;
    %(type)s* gpu_%(name)s;
    long int %(name)s_device_to_host;
    bool %(name)s_host_to_device;
"""
}

attribute_cpp_init = {
    'local': """
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(%(init)s);
        if (%(name)s.empty())
            return false;
        gpu_%(name)s = init_matrix_variable_gpu<%(type)s>(%(name)s);
        if (gpu_%(name)s == nullptr)
            return false;

        %(name)s_host_to_device = true;
        %(name)s_device_to_host = t;
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(%(init)s);
        gpu_%(name)s = init_vector_variable_gpu<%(type)s>(%(name)s);
        %(name)s_dirty = true;
""",
    'global': """
        // Global %(attr_type)s %(name)s
        %(name)s = static_cast<%(type)s>(%(init)s);
        cudaMalloc((void**)&gpu_%(name)s, sizeof(%(type)s));
        %(name)s_dirty = true;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(bool);
        size_in_bytes += sizeof(%(ctype)s*);
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        cudaFree(gpu_%(name)s);
""",
    'semiglobal': """
        // %(name)s
        cudaFree(gpu_%(name)s);
""",
    'global': ""
}

attribute_host_to_device = {
    'local': """
        // %(name)s: local
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), this->tile_data_.size() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, %(name)s.data(), this->nb_dendrites() * sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_host_to_device )
        {
        #ifdef _DEBUG
            std::cout << "HtoD: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( gpu_%(name)s, &%(name)s, sizeof( %(type)s ), cudaMemcpyHostToDevice);
            %(name)s_host_to_device = false;
        #ifdef _DEBUG
            cudaError_t err = cudaGetLastError();
            if ( err!= cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err) << std::endl;
        #endif
        }
"""
}

attribute_device_to_host = {
    'local': """
        // %(name)s: local
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, this->tile_data_.size() * sizeof( %(type)s ), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'semiglobal': """
        // %(name)s: semiglobal
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( %(name)s.data(), gpu_%(name)s, this->nb_dendrites() * sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
""",
    'global': """
        // %(name)s: global
        if ( %(name)s_device_to_host < t ) {
        #ifdef _DEBUG
            std::cout << "DtoH: %(name)s ( proj%(id)s )" << std::endl;
        #endif
            cudaMemcpy( &%(name)s, gpu_%(name)s, sizeof(%(type)s), cudaMemcpyDeviceToHost);
        #ifdef _DEBUG
            cudaError_t err_%(name)s = cudaGetLastError();
            if ( err_%(name)s != cudaSuccess )
                std::cout << "  error: " << cudaGetErrorString(err_%(name)s) << std::endl;
        #endif
            %(name)s_device_to_host = t;
        }
"""
}

# BSR implementation following Eberhardt & Hoemmen (2016) - row-per-thread
#
# In this variant, each thread computes one row in a dense block and is intended for larger block sizes.
# We ensure, that the number of threads in a CUDA block is equal to the tile-size. Further we assume
# squared dense blocks. Last but not least, each CUDA block computes at least one blocked row in the BSR.
#
rate_psp_kernel_rpt = {
    'device_kernel': {
        'sum': """
__global__ void cu_proj%(id_proj)s_psp_bsr(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s ) {
    const %(idx_type)s idx = threadIdx.x;
    const %(idx_type)s block_row = blockIdx.x;
    const %(size_type)s tile_size2 = tile_size * tile_size;
    extern %(float_prec)s __shared__ loc_pre_r[];

    for (%(idx_type)s row = block_row; row < n_block_rows; row += gridDim.x) {
        const %(idx_type)s first_block = row_ptr[row];
        const %(idx_type)s last_block = row_ptr[row + 1];

        %(float_prec)s lsum = 0.0;
        for (%(idx_type)s block = first_block; block < last_block; block++)
        {
            __syncthreads();
            loc_pre_r[idx] = pre_r[col_ids[block] * tile_size +idx];
            __syncthreads();

            const %(size_type)s tile_off = block * tile_size2;
            for (%(idx_type)s col = 0; col < tile_size; col++)
                lsum += %(psp)s
        }

        %(target_arg)s[row * tile_size + idx] += lsum;
    }
}
"""
    },
    'device_header': """__global__ void cu_proj%(id)s_psp_bsr(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s );
""",
    'host_call': """
    // proj%(id_proj)s: pop%(id_pre)s -> pop%(id_post)s
    if ( pop%(id_post)s->_active && proj%(id_proj)s->_transmission ) {
        unsigned int nb_blocks = std::min<unsigned int>(proj%(id_proj)s.block_row_size(), 65535);
        size_t smem_size = proj%(id_proj)s->get_tile_size() * sizeof(%(float_prec)s);
        cu_proj%(id_proj)s_psp_bsr<<<nb_blocks, proj%(id_proj)s->get_tile_size(), smem_size>>>(
            %(conn_args)s
            /* other variables */
            %(add_args)s
            /* result */
            %(target_arg)s
        );

    #ifdef _DEBUG
        auto err = cudaGetLastError();
        if ( err != cudaSuccess ) {
            std::cout << "cu_proj%(id_proj)s_psp: " << cudaGetErrorString(err) << std::endl;
        }
    #endif
    }
""",
    'thread_init': {
        'float': {
            'sum': "0.0f",
            'min': "FLT_MAX",
            'max': "FLT_MIN",
            'mean': "0.0f"
        },
        'double': {
            'sum': "0.0",
            'min': "DBL_MAX",
            'max': "DBL_MIN",
            'mean': "0.0"
        }
    }
}

# BSR implementation following Eberhardt & Hoemmen (2016) - column-by-column
#
# This variant is intended for small block sizes.
#
# Keys - structure:
#
# * device_kernel
#    * sum
# * kernel_decl
# * host_call
# * thread_init
rate_psp_kernel_cbc = {
    'device_kernel': {
        'sum': """
__global__ void cu_proj%(id_proj)s_psp(%(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s ) {
    unsigned int tIdx = threadIdx.x;
    unsigned int bIdx = blockIdx.x;
    extern %(float_prec)s __shared__ sdata[];
    %(idx_type)s tile_size2 = tile_size *  tile_size;

    int blocks_per_warp = floorf(blockDim.x / tile_size2);
    int block_offset = int(float(tIdx) / float(tile_size2));

    // reset
    sdata[tIdx] = 0;

    // iterate across all columns in this row (determined by blockIdx)
    for (%(idx_type)s col_idx = row_ptr[bIdx]+block_offset; col_idx < row_ptr[bIdx+1]; col_idx+=blocks_per_warp) {
        %(idx_type)s dense_col_idx = (tIdx / tile_size) %% tile_size;
        %(idx_type)s dense_val_idx = tIdx %% tile_size2;

        // which dense column, determine where to access pr
        int bcol_idx = col_ids[col_idx];
        
        // perform dense SpMV (column_major)
        const %(float_prec)s* loc_values = w + col_idx * tile_size2;
        const   %(float_prec)s* loc_pr = pre_r + bcol_idx * tile_size;

        sdata[tIdx] += loc_values[dense_val_idx] * loc_pr[dense_col_idx];
    }

    // reduction to first tile
    if (tIdx < tile_size2) {
        for(%(idx_type)s i = tIdx+tile_size2; i < blockDim.x; i+= tile_size2) {
            sdata[tIdx] += sdata[i];
        }
    }

    // reduction within tile
    // and write back result
    if (tIdx < tile_size) {
        for (int i = tIdx; i < tile_size2; i+= tile_size) {
            %(target_arg)s[bIdx * tile_size+tIdx] += sdata[i];
        }
    }
}
""",
},
    'invoke_kernel': """
void call_proj%(id_proj)s_psp(RunConfig cfg, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s ) {
    // kernel launch
    cu_proj%(id_proj)s_psp<<<cfg.nb, cfg.tpb, cfg.smem_size, cfg.stream>>>(
        %(conn_args_call)s
        /* other variables */
        %(add_args_call)s
        /* result */
        %(target_arg_call)s
    );
}
""",
    'kernel_decl': """void call_proj%(id_proj)s_psp(RunConfig cfg, %(conn_args)s%(add_args)s, %(float_prec)s* %(target_arg)s );
""",
    'host_call': """
    // proj%(id_proj)s: pop%(id_pre)s -> pop%(id_post)s
    if ( pop%(id_post)s->_active && proj%(id_proj)s->_transmission ) {
        // kernel configuration
        RunConfig proj%(id_proj)s_psp_cfg;
        proj%(id_proj)s_psp_cfg.nb = proj%(id_proj)s->_nb_blocks;
        proj%(id_proj)s_psp_cfg.tpb = proj%(id_proj)s->_threads_per_block;
        proj%(id_proj)s_psp_cfg.smem_size = proj%(id_proj)s->_threads_per_block * sizeof(%(float_prec)s);    // one local variable per thread
        proj%(id_proj)s_psp_cfg.stream = proj%(id_proj)s->stream;

        // invoke kernel
        call_proj%(id_proj)s_psp (
            /* kernel config */
            proj%(id_proj)s_psp_cfg,
            /* ranks and offsets */
            %(conn_args)s
            /* computation data */
            %(add_args)s
            /* result */
            %(target_arg)s
        );

    #ifdef _DEBUG
        auto err = cudaGetLastError();
        if ( err != cudaSuccess ) {
            std::cout << "cu_proj%(id_proj)s_psp: " << cudaGetErrorString(err) << std::endl;
        }
    #endif
    }

""",
    'thread_init': {
        'float': {
            'sum': "0.0f",
            'min': "FLT_MAX",
            'max': "FLT_MIN",
            'mean': "0.0f"
        },
        'double': {
            'sum': "0.0",
            'min': "DBL_MAX",
            'max': "DBL_MIN",
            'mean': "0.0"
        }
    }
}

conn_templates = {
    # connectivity representation
    'conn_header': "const %(idx_type)s* __restrict__ row_ptr, const %(idx_type)s* __restrict__ col_ids, const %(idx_type)s n_block_rows, const %(idx_type)s tile_size",
    'conn_call': "proj%(id_proj)s->gpu_block_row_pointer(), proj%(id_proj)s->gpu_block_column_index(), proj%(id_proj)s->block_row_size(), proj%(id_proj)s->get_tile_size()",
    'conn_kernel': "row_ptr, col_ids, n_block_rows, tile_size",

    # launch config
    'launch_config': launch_config,

    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'host_to_device': attribute_host_to_device,
    'device_to_host': attribute_device_to_host,

    # operations
    'rate_psp': rate_psp_kernel_cbc,
    #'rate_psp': rate_psp_kernel_rpt,
}

conn_ids = {
    'local_index': "[tile_off + col * tile_size + idx]",
    'semiglobal_index': '[i]',
    'global_index': '[0]',
    'pre_index': '[first_col_x + col]',
    'post_index': '[row_indices[j]]',
    'pre_prefix': 'pre_',
    'post_prefix': 'post_',
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();       
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

#############################################
##  Synaptic delay
#############################################
delay = {
    'uniform': {
        'declare': """
    // Uniform delay
    int delay ;""",

        'pyx_struct':
"""
        # Uniform delay
        int delay""",
        'init': """
    delay = delays[0][0];
""",
        'pyx_wrapper_init':
"""
        proj%(id_proj)s.delay = syn.uniform_delay""",
        'pyx_wrapper_accessor':
"""
    # Access to non-uniform delay
    def get_delay(self):
        return proj%(id_proj)s.delay
    def get_dendrite_delay(self, idx):
        return proj%(id_proj)s.delay
    def set_delay(self, value):
        proj%(id_proj)s.delay = value
"""
    },
    'nonuniform_rate_coded': None,
    'nonuniform_spiking': None
}

###############################################################
# Rate-coded continuous transmission
###############################################################
continuous_transmission = {
    'sum' : """
%(pre_copy)s

// iterate across all blocks
%(omp_code)s %(omp_clause)s %(omp_schedule)s
for (%(idx_type)s i = 0; i < num_blocks_; i++) {
    
    //compute maxlength (maxnzr) in each block
    %(size_type)s maxlength = (row_ptr_[i + 1] - row_ptr_[i]) / block_size_;

    // iterate over all values within the block
    for (%(idx_type)s j = 0; j < block_size_; j++) {
        %(size_type)s pos = row_ptr_[i] + j * maxlength;
        //row_now: compute global row idx
        %(size_type)s row_now = i * block_size_ + j;

        if (row_now < num_rows_) {
            %(float_prec)s sum = 0.0;
            for (int k = 0; k < maxlength; k++) {
                %(idx_type)s rk_pre = col_idx_[pos+k];

                sum += %(psp)s;
            }
            pop%(id_post)s->_sum_%(target)s[row_now] += sum;
        }
        
    }
}
""",
    'max': "",
    'min': "",
    'mean': "",
}

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,
    
    # operations
    'rate_coded_sum': continuous_transmission,
    'vectorized_default_psp': {},
    'update_variables': None
}

conn_ids = {
    'local_index': '[pos+k]',
    'semiglobal_index': '[i]',
    'global_index': '',
    'post_index': '[rk_post]',
    'pre_index': '[rk_pre]',
    'delay_u' : '[delay-1]' # uniform delay
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Neuron import Neuron

available_read_outs = [
    "spike_count",
    "time_to_first_spike",
    "time_to_k_spikes",
    "membrane_potential"
]

# Default neuron used in hidden layers
IaF = Neuron(
    parameters = """
        vt = 1          : population
        vr = 0          : population
        mask_tau = 20.0 : population
    """,
    equations = """
        dv/dt    = g_exc          : init = 0.0 , min=-2.0
        dmask/dt = -mask/mask_tau : init = 0.0
    """,
    spike = """
        v > vt
    """,
    reset = """
        v = vr
        mask += 1.0/mask_tau
    """,
    name = "IaF neuron"
)

# Default neuron used in read-out layer
IaF_ReadOut = Neuron(
    parameters = """
        vt = 1          : population
        vr = 0          : population
    """,
    equations = """
        dv/dt = g_exc   : init = 0.0 , min=-2.0
    """,
    spike = """
        v > vt
    """,
    reset = """
        v = vr
    """,
    name = "IaF neuron"
)

# Used as output layer for "time_to_k_spikes" read-out.
IaF_TTKS = Neuron(
    parameters = """
        k = 0           : population
        vt = 1          : population
        vr = 0          : population
        sc = 0
    """,
    equations = """
        dv/dt = g_exc   : init = 0.0 , min=-2.0
    """,
    spike = """
        v > vt
    """,
    reset = """
        v = vr
        sc += 1
    """
)

# Used as output layer for "max_membrane_potential" read-out.
IaF_Acc = Neuron(
    equations = """
        dv/dt = g_exc   : init = 0.0 , min=-2.0
    """,
    spike = """
        v < -10000 # will never happen
    """,
    reset = ""
)
